#!/usr/bin/python3
#
# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
#
"""
Contains the parent "Backend" class that is inherited by all backend modules.

A "backend" implements a data layer of processing between Polychromatic's interfaces
and the vendor's driver, daemon or other implementation. This application acts like
an orchestrator for presenting and relaying instructions, but is not expected to
actually send binary or payloads to the hardware itself.

Refer to the online documentation for more details:
https://docs.polychromatic.app/
"""

from ..fx import FX
import glob
import os
import grp


class BackendBase(object):
    """
    All backends inherit from this class. Contains useful functions and any
    reimplementations from Polychromatic's base class.
    """
    def __init__(self, base):
        self.backend_id = "Unknown"
        self._base = base

        # Pass a function for translating strings in UI messages. Ignore for debug() messages.
        self._ = base._

    def debug(self, message=""):
        """
        Use this function to output messages to the user when they have verbose enabled.
        This may be useful when users are diagnosing issues.
        """
        self._base.dbg.stdout("[{0}] {1}".format(self.backend_id, str(message)), self._base.dbg.debug, 1)

    def get_backend_storage_path(self):
        """
        Returns a path for storing data for the backend.
        """
        config_path = os.path.join(self._base.paths.config, "backends", self.backend_id)
        if not os.path.exists(config_path):
            os.makedirs(config_path)
        return config_path

    def get_form_factor(self, form_factor="unrecognised"):
        """
        Return a form factor dictionary for Polychromatic.

        See common.get_form_factor() for the dictionary output.
        See common.FORM_FACTORS for a list of valid IDs to pass to this function.
        """
        return self._base.common.get_form_factor(self._, form_factor)

    def get_icon(self, folder="", name=""):
        """
        Return an icon path from: data/img/<folder>/<name>.svg
        The file extension is omitted.
        """
        return self._base.common.get_icon(folder, name)

    def get_exception_as_string(self, e):
        """
        Returns a traceback in string format. Use this to relay error messages to the user.

        try:
            <something that might fail>
        except Exception as e:
            return self.get_exception_as_string(e)
        """
        return self._base.common.get_exception_as_string(e)

    def __repr__(self):
        return self.backend_id


class Backend(BackendBase):
    """
    A backend implementing the communication required between this software
    and a driver/daemon.

    Most functions are stubs and must be reimplemented.
    """
    def __init__(self, *args):
        super().__init__(*args)

        # Set the Backend ID here
        # (Also to be added in ../middleman.py)
        self.backend_id = "unknown"

        # Name of backend
        self.name = "OpenRazer"

        # Filename of the logo stored in data/img/logo/ (SVG preferred)
        self.logo = "example.svg"

        # Set this string to the backend version
        self.version = "0.1.0"

        # URLs and license for the backend
        self.project_url = ""
        self.bug_url = ""
        self.releases_url = ""
        self.license = "GPLv3"

        # This module may contain useful functions. See BackendHelpers() for usage.
        self.helpers = BackendHelpers()

    def init(self):
        """
        Perform the logic for initalizing the backend, such as connecting
        to a daemon using the necessary library.

        Return:
            - True      Success!
            - (str)     Traceback/error message. Cannot use backend.
        """
        raise NotImplementedError

    class UnknownDeviceItem(object):
        """
        An object describing a device that may potentially be compatible, but
        in its current state, cannot be used.

        For example, a installation problem or permission error prevents
        this device from being controlled.
        """
        def __init__(self):
            # Human readable name of the device (if known)
            self.name = "Unknown"

            # This is assigned automatically.
            self.backend = Backend

            # VID:PID of the device
            self.vid = ""
            self.pid = ""

            # Use Backend.get_form_factor(), passing an ID from common.FORM_FACTORS
            # that identifies this device. Pass "unrecognised" to function if unknown.
            self.form_factor = {}

            # Indicates whether a device is simply unusable due to an error or unsupported
            self.supported = False

    def get_unsupported_devices(self):
        """
        Returns a list of UnknownDeviceItem(), or empty list.
        """
        return []

    #####################################################################
    class DeviceItem(object):
        """
        An object describing the complete state of the device and its available
        functions. This may include current settings, device options,
        serial number and firmware version.

        This object also contains the code for executing options and parameters,
        as well as defining how Polychromatic should present them.
        """
        def __init__(self):
            # Human readable name of the device (including vendor name)
            self.name = "Unnamed Device"

            # This is assigned automatically.
            self.backend = Backend

            # Use Backend.get_form_factor(), passing an ID from common.FORM_FACTORS
            self.form_factor = {}

            # Local or remote file path to a 'real' image of the hardware.
            # (Leave empty string for no image)
            self.real_image = ""

            # String containing the device's serial number, must be unique.
            self.serial = "X"

            # Does this device only have one colour?
            # Set to True for hardware that have individually addressable LEDs,
            # but physically only displays one colour from the RGB range.
            self.monochromatic = False

            # Device's vendor and product ID
            self.vid = "????"
            self.pid = "????"

            # If applicable, a string describing the firmware version, e.g. "v1.0"
            self.firmware_version = ""

            # If applicable, a string describing the keyboard locale, e.g. "en_GB"
            # This will be used for determining graphics in the effect editor.
            self.keyboard_layout = ""

            # Is this device marketed as having programmable keys via software?
            self.has_programmable_keys = False

            # Is this device marketed as having macro recording? (deprecated)
            self.has_macro_keys = False

            # Stores a Battery() object, if the device has a battery.
            self.battery = None

            # Stores a DPI() object, unless device does not support DPI X/Y.
            self.dpi = None

            # Stores a Matrix() object, if device supports per-LED lighting.
            self.matrix = None

            # List of Zone() objects.
            self.zones = []

        def __str__(self):
            return self.name

        def __repr__(self):
            return "{0}:{1}".format(self.serial, self.name.replace(" ", ""))

        def refresh(self):
            """
            This function is called before showing the current status for a device, such as:
            - Controller: After selecting/refreshing a device in the Device tab
            - Tray: Once when the applet starts
            - CLI: Listing the current device status

            It is expected the device's options and features have their correct values at this
            point, and acts as a cache until the device is refreshed again. It co-exists alongside
            each option's .refresh() function, which is used when the application needs
            to refresh a specific option or feature.

            Each backend may implement this differently depending whether the
            hardware knows what it's up to or if it uses a software persistence implementation.
            """
            return

        class Battery(object):
            """
            An object storing battery information with a function for refreshing.
            """
            def __init__(self):
                self.is_charging = False
                self.percentage = -1
                self.is_removable = False

            def refresh(self):
                """
                Update the battery status.
                """
                raise NotImplementedError

        class DPI(object):
            """
            An object storing the current DPI values and get/set functions.
            """
            def __init__(self):
                self.x = 0
                self.y = 0
                self.min = 0
                self.max = 0
                self.default_stages = []
                self.can_sync = False

                # Populated by Polychromatic later, if the user saved any custom DPI stages
                self.user_stages = []

            def refresh(self):
                """
                Reload the DPI variables stored in this object.
                """
                raise NotImplementedError

            def set(self, x, y):
                """
                Sets the new DPI value to the specified X/Y value.
                """
                raise NotImplementedError

            def sync(self, stages):
                """
                Synchronises DPI stages to the hardware buttons when 'can_sync' is true.

                "stages" is a list consisting of a sublist of X,Y values, e.g.
                [ [1800, 1800], [6500, 4800], [...] ]
                """
                raise NotImplementedError

        class Matrix(FX):
            """
            An object holding data and objects for individual LED software-driven lighting,
            if supported by the device.

            Devices with NAND/flash memory that isn't designed for repeated usage
            should not implement this feature, as it may damage the hardware.

            See also: fx.FX()
            """
            def __init__(self):
                self.name = "Unknown Device"
                self.form_factor_id = "unrecognised"
                self.rows = 0
                self.cols = 0

            def init(self):
                """
                Prepare the device for custom frames. If unnecessary, this can be ignored.
                """
                return

            def set(self, x=0, y=0, red=255, green=255, blue=255):
                """
                Set a colour at the specified co-ordinate.
                """
                raise NotImplementedError

            def draw(self):
                """
                Send the data to the hardware to be displayed.
                """
                raise NotImplementedError

            def clear(self):
                """
                Reset all LEDs to an off state.
                """
                raise NotImplementedError

            def brightness(self, percent):
                """
                Set the global brightness of the LEDs. Could be used for fade effects globally.
                """
                raise NotImplementedError

        class Zone(object):
            """
            An object that describes a specific lighting area of the hardware.
            If the device has no concept of this or is monolithic, call this zone "main".
            """
            def __init__(self):
                # Internal ID
                self.zone_id = ""

                # Human readable text describing this zone, e.g. "Left Side"
                self.label = "Unknown Zone"

                # Full path using self.get_icon() - usually from {data}/img/zones/
                self.icon = ""

                # List of Option() objects - see below.
                self.options = []

                def __str__(self):
                    return self.zone_id

                def __repr__(self):
                    return self.zone_id

    class Option(object):
        """
        Options are settings that the user can change. This is the base class,
        use one of the child classes below. These tell Polychromatic how to
        present them, the parameters (if any) and the function to execute.
        """
        def __init__(self):
            # Internal to identify this option later.
            self.uid = ""

            # Human readable text describing this option, e.g. "Brightness"
            self.label = "Unknown Option"

            # Full path using self.get_icon() - usually from {data}/img/options/
            self.icon = ""

            # ------ The following depend on the option ------
            # Is this option currently selected?
            self.active = False

            # List of Parameter() objects
            self.parameters = []

            # Does selecting this option need a colour?
            self.colours_required = 0

            # Which colours are assigned for this option?
            #   - Initially, this should be populated with previously set colours.
            #   - Values in the list change when the user changes colours via the interface.
            #   - Must be same length as colours_required.
            # Format: ["#RRGGBB"]
            self.colours = []

        def __str__(self):
            return self.uid

        def __repr__(self):
            return self.uid

        def refresh(self):
            """
            Refresh the active variable for the option and that of any parameters.
            If the option doesn't have an active state, ignore this function.

            Nothing is returned from this function. However, if an exception occurs
            or is manually raised, an error will be shown to the user.
            """
            return

        def apply(self, data=None):
            """
            Execute the action on the device. The "data" argument varies by option type.

            Nothing is returned from this function. However, if an exception occurs
            or is manually raised, an error will be shown to the user.
            """
            raise NotImplementedError

        class Parameter(object):
            """
            An object describing a parameter that can be selected. For example, a
            "Wave" option may have a "Fast" and "Slow" mode.

            This Parameter() object is passed to an option's apply() function.
            """
            def __init__(self):
                # Can be any data type, this will be passed as an argument to Option.apply()
                self.data = None

                # Human readable text describing this option, e.g. "Fast"
                self.label = "Unknown Parameter"

                # Full path using get_icon() - usually from {data}/img/params/
                self.icon = ""

                # Is this parameter currently selected?
                self.active = False

                # Select this parameter as a fallback?
                # (Only one should be default. If there's no default, use the first one)
                self.default = False

                # Does selecting this parameter require a colour?
                # (Make sure the option's colour_list is populated)
                self.colours_required = 0

            def __str__(self):
                return str(self.data)

            def __int__(self):
                return int(self.data)

            def __repr__(self):
                return str(self.data)

    class EffectOption(Option):
        """
        For presenting hardware effects. These are grouped together
        under one section/menu. There are no additional variables to set.

        Parameters: Optional
        Colours: Optional
        """
        def __init__(self):
            super().__init__()

    class ToggleOption(Option):
        """
        For options that are either on or off.

        Parameters: Ignored
        Colours: Ignored
        """
        def __init__(self):
            super().__init__()

            # Optionally change the labels depending on interface (menu/checkbox)
            self.label_enable = "" # Enable
            self.label_disable = "" # Disable
            self.label_toggle = "" # Enabled

            # Optionally the tray can show alternate icons representing on/off states
            self.icon_enable = ""
            self.icon_disable = ""

        def apply(self, enabled=True):
            """
            Execute the action on the device. This argument will be a boolean.
            """
            raise NotImplementedError

    class SliderOption(Option):
        """
        For an option that is a variable between two integers.

        Parameters: Ignored
        Colours: Ignored
        """
        def __init__(self):
            super().__init__()

            # Current value, the range, how much to step and the suffix strings
            self.value = 0
            self.min = 1
            self.max = 100
            self.step = 1
            self.suffix = ""
            self.suffix_plural = ""

        def apply(self, value=0):
            """
            Execute the action on the device. This argument will be an integer.
            """
            raise NotImplementedError

    class MultipleChoiceOption(Option):
        """
        For an option that should be presented as a drop down or list.
        There are no variables to set. These are populated from parameters.

        Parameters: Required
        Colours: Ignored
        """
        def __init__(self):
            super().__init__()

    def get_devices(self):
        """
        Return:
            - (list)  A list of DeviceItem() objects.
            - (str)   Traceback/error message. Cannot continue.
        """
        return NotImplementedError

    def get_device_by_name(self, name):
        """
        For the application to quickly retrieve a device object based on the
        name of the device.

        Return:
            - DeviceItem()      Found the specified device.
            - (str)             Traceback/error message. Cannot use device.
            - None              Device not found.
        """

    def get_device_by_serial(self, serial):
        """
        For the application to quickly retrieve a device object based on the
        serial number.

        Return:
            - DeviceItem()      Found the specified device.
            - (str)             Traceback/error message. Cannot use device.
            - None              Device not found.
        """

    def troubleshoot(self, fn_progress_set_max, fn_progress_advance):
        """
        Perform troubleshooting steps to identify issues with the installation of
        the backend. These checks could include verifying the device is being
        detected, or a binary is accessible (/usr/bin/xyz), for instance.

        If the backend is simple in nature, implementing a troubleshooter might not be necessary.

        fn_progress_set_max and fn_progress_advance are functions passed to this function.
        To optionally provide feedback using the progress bar, you can call fn_progress_set_max(int)
        with the maximium value. Then, call fn_progress_advance() to add 1 to the progress bar.

        Troubleshooting code should be implemented in polychromatic/troubleshoot/<backend>.py.

        Return:
            - (list)  A list of dictionary results in format below.
            - (str)   Traceback/error message. Cannot continue.
            - None    Troubleshooter not avaliable on this operating system

        Expected data:
        [
            {
                "test_name":    (str)   i18n enabled string describing the test
                "suggestion":   (str)   i18n enabled string describing what to do on failure.
                "passed":       (bool)  No problems found. None if undetermined.
            },
            {...}
        ]
        """
        return None

    def restart(self):
        """
        User requests to restart the backend - which could be a daemon, or
        other command to reinitialise the devices.

        For some backends, this may not even be necessary.

        Return:
            - True      Successfully executed restart.
            - False     Failed to restart.
            - None      Not applicable.
        """
        return None


class BackendHelpers():
    """
    Shared functions that are useful for backends.
    """
    def get_usb_pids_by_vid(self, vid_to_find):
        """
        Returns a integer list of USB PIDs for a VID plugged into the system.
        """
        vendor_files = glob.glob("/sys/bus/usb/devices/*/idVendor")
        found_pids = []
        for vendor in vendor_files:
            with open(vendor, "r") as f:
                vid = str(f.read()).strip().upper()
                if vid == vid_to_find:
                    with open(os.path.dirname(vendor) + "/idProduct") as f:
                        pid = str(f.read()).strip().upper()
                        found_pids.append(pid)
        return found_pids

    def is_user_in_group(self, group):
        """
        Check the user groups for the currently logged in user and returns a
        boolean to indicate whether the specified group was found.
        """
        if group in [grp.getgrgid(g).gr_name for g in os.getgroups()]:
            return True

        return False



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
This module contains widgets shared across the Controller GUI.
"""

from ..base import PolychromaticBase
from .. import common
from .. import locales
from .. import preferences as pref
from .. import procpid
from .. import fileman

from ..qt.flowlayout import FlowLayout as QFlowLayout

import hashlib
import os
import glob
import requests
import shutil

from PyQt5 import uic, QtSvg
from PyQt5.QtCore import Qt, QSize, QMargins
from PyQt5.QtGui import QIcon, QPalette, QColor, QFont, QPixmap, QMovie
from PyQt5.QtSvg import QSvgWidget
from PyQt5.QtWebEngineWidgets import QWebEngineView
from PyQt5.QtWidgets import QApplication, QMainWindow, QMenuBar, \
                            QWidget, QMessageBox, QGridLayout, \
                            QLabel, QPushButton, QToolButton, QGroupBox, \
                            QListWidget, QHBoxLayout, QVBoxLayout, QFormLayout, \
                            QSizePolicy, QSpacerItem, QDialog, QColorDialog, \
                            QDialogButtonBox, QTreeWidget, QTreeWidgetItem, \
                            QLineEdit, QTextEdit, QTabWidget, QScrollArea, \
                            QButtonGroup, QFileDialog, QMenu, QAction, \
                            QDockWidget, QCheckBox, QSpinBox, QComboBox, \
                            QTreeWidget, QDoubleSpinBox, QRadioButton, QToolBar


def load_qt_theme(appdata, window):
    """
    Apply the Polychromatic Qt theme for the main window if enabled by the user.

    Params:
        appdata         ApplicationData() object
        window          QMainWindow() or QDialog() object
    """
    if appdata.system_qt_theme:
        return

    # Load "Play" font
    font = QFont("Play", 10, 0)
    window.setFont(font)

    menu_bar = window.findChild(QMenuBar)
    if menu_bar:
        menu_bar.setFont(font)

    custom_tabs = window.findChild(QWidget, "MainTabCustom")
    if custom_tabs:
        custom_tabs.setFont(font)

    # Load basic colour palettes
    window.setPalette(get_palette(appdata))

    # Load QSS (essentially CSS) with Polychromatic's design
    with open(os.path.join(common.paths.data_dir, "qt", "style.qss"), "r") as f:
        window.setStyleSheet(f.read().replace("[data]", common.paths.data_dir))


def get_palette(app):
    """
    Returns a QPalette with Polychromatic's colours.
    """
    palette = QPalette()
    black = QColor(0, 0, 0)
    white = QColor(255, 255, 255)
    primary = QColor(0, 255, 0)
    secondary = QColor(0, 128, 0)
    palette.setColor(QPalette.Window, QColor(0, 0, 0))
    palette.setColor(QPalette.WindowText, white)
    palette.setColor(QPalette.Base, QColor(25, 25, 25))
    palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
    palette.setColor(QPalette.ToolTipBase, white)
    palette.setColor(QPalette.ToolTipText, white)
    palette.setColor(QPalette.Text, white)
    palette.setColor(QPalette.Button, QColor(50, 50, 50))
    palette.setColor(QPalette.ButtonText, white)
    palette.setColor(QPalette.Link, primary)
    palette.setColor(QPalette.Highlight, secondary)
    palette.setColor(QPalette.HighlightedText, white)
    return palette


def get_ui_widget(appdata, name, q_toplevel=QWidget):
    """
    Returns a QWidget object for the specified .ui file.

    Params:
        appdata         ApplicationData() object
        name            Name of UI file (in ui/ folder) without .ui extension
        q_toplevel      Top-level object class, e.g. QWidget() or QDialog()
    """
    ui_file = os.path.join(appdata.paths.data_dir, "qt", name + ".ui")
    if not os.path.exists(ui_file):
        print("Missing UI file: " + ui_file)
        return None

    # HACK: Allows VS Code IDE to autocomplete, since PyQt5-stubs haven't stubbed uic yet
    # https://github.com/python-qt-tools/PyQt5-stubs/issues/150
    if 1 == 2:
        return QWidget()

    widget = uic.loadUi(ui_file, q_toplevel())

    # Apply the styles for dialogs/windows
    if q_toplevel in [QDialog, QMainWindow]:
        load_qt_theme(appdata, widget)

    # Process i18n strings for existing widgets
    translate_ui(appdata, widget)

    # Combo boxes render a light line at the top/bottom under light themes
    # Does not happen when combo boxes are created during runtime
    if not appdata.system_qt_theme:
        for combobox in widget.findChildren(QComboBox):
            combobox.view().parentWidget().setStyleSheet('background: transparent')

    # Theme icons inside a QDialogButtonBox widget
    dialog_buttons = widget.findChild(QDialogButtonBox)
    if dialog_buttons:
        PolychromaticWidgets(appdata).set_dialog_buttons_icons(dialog_buttons)

    return widget


def translate_ui(appdata, widget):
    """
    Iterates over all the translatable widgets from the newly loaded .ui file
    and use gettext to apply the localized strings.
    """
    if appdata.locales.get_current_locale() == "en_GB":
        return

    for widget_type in [QLabel, QMenu, QAction, QPushButton, QToolButton, \
                        QTabWidget, QTreeWidgetItem, QDockWidget, QTreeWidget, \
                        QCheckBox, QRadioButton, QSpinBox, QDoubleSpinBox, \
                        QComboBox, QLineEdit, QGroupBox]:
        children = widget.findChildren(widget_type)
        for subwidget in children:
            _translate_widget(appdata, subwidget)

    _ = appdata._

    if type(widget) == QDialog:
        widget.setWindowTitle(_(widget.windowTitle()))


def _translate_widget(appdata, widget):
    """
    Translates the strings of a widget loaded using uic.loadUi()
    """
    _ = appdata._

    def _translate(set_function, string: str, add_space=False):
        if not string:
            return

        new_string = _(string)
        if string == new_string:
            new_string = _(string.strip())
        if add_space:
            new_string = " " + new_string
        if string != new_string:
            set_function(_(new_string))

    if type(widget) == QWidget:
        return

    if type(widget) in [QMenu, QGroupBox]:
        return _translate(widget.setTitle, widget.title())

    if type(widget) == QDockWidget:
        return _translate(widget.setWindowTitle, widget.windowTitle())

    if type(widget) in [QSpinBox, QDoubleSpinBox]:
        _translate(widget.setPrefix, widget.prefix())
        _translate(widget.setSuffix, widget.suffix(), add_space=True)
        _translate(widget.setToolTip, widget.toolTip())
        _translate(widget.setStatusTip, widget.statusTip())
        return

    if type(widget) == QTabWidget:
        for index in range(0, widget.count()):
            if not widget.tabText(index) == "":
                widget.setTabText(index, _(widget.tabText(index)))
            if not widget.tabToolTip(index) == "":
                widget.setTabToolTip(index, _(widget.tabToolTip(index)))
        return

    if type(widget) == QComboBox:
        for index in range(0, widget.count()):
            if not widget.itemText(index) == "":
                widget.setItemText(index, _(widget.itemText(index)))
        return

    if type(widget) == QTreeWidget:
        tree_root = widget.invisibleRootItem()
        for index in range(0, tree_root.childCount()):
            item = tree_root.child(index)
            for column in range(0, item.columnCount()):
                if not item.text(column) == "":
                    item.setText(column, _(item.text(column)))

            for subindex in range(0, item.childCount()):
                child = item.child(subindex)
                for column in range(0, item.columnCount()):
                    if not child.text(subindex) == "":
                        child.setText(column, _(child.text(subindex)))
        return

    # All other UI controls: QCheckBox, QLabel, QRadioButton, QPushButton, etc
    _translate(widget.setText, widget.text())
    _translate(widget.setToolTip, widget.toolTip())
    _translate(widget.setStatusTip, widget.statusTip())


def clear_layout(layout):
    """
    Removes all Qt elements inside a layout.
    """
    while layout.count():
        child = layout.takeAt(0)
        if child.widget():
            child.widget().deleteLater()


def set_pixmap_for_label(qlabel, icon_path, icon_size=24):
    """
    Creates a pixmap for a label maintaining its aspect ratio and size.

    This only applies to QLabel() objects. Buttons and other UI elements use QIcon().

    Params:
        qlabel      (obj)   QLabel() object
        icon_path   (str)   Absolute path to icon
        icon_size   (int)   Dimensions to scale
    """
    pixmap_src = QPixmap(icon_path)
    pixmap = pixmap_src.scaled(icon_size, icon_size, Qt.KeepAspectRatio, Qt.SmoothTransformation)
    qlabel.setPixmap(pixmap)
    qlabel.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)


def get_real_device_image(uri):
    """
    Returns the path to the image that physically represents a device. If the
    URI is HTTP(S), this will be downloaded and cached locally.
    """
    if os.path.exists(uri):
        return uri

    cache_key = hashlib.md5(uri.encode("UTF-8")).hexdigest()
    cache_path = os.path.join(PolychromaticBase.paths.devices, cache_key)

    if os.path.exists(cache_path):
        return cache_path

    if not PolychromaticBase.preferences.get("controller", {}).get("download_device_images", True):
        return ""

    if uri.startswith("http://") or uri.startswith("https://"):
        PolychromaticBase.dbg.stdout("Downloading image: " + uri, PolychromaticBase.dbg.action, 1)
        try:
            r = requests.get(uri)
        except requests.exceptions.ConnectionError as e:
            PolychromaticBase.dbg.stdout(f"Connection error while downloading: {uri}\n{str(e)}", PolychromaticBase.dbg.warning)
            return ""
        if r.status_code == 200:
            PolychromaticBase.dbg.stdout("Retrieved image: " + uri, PolychromaticBase.dbg.action, 1)
            with open(cache_path, "wb") as f:
                f.write(r.content)
            return cache_path
        else:
            PolychromaticBase.dbg.stdout(f"Connection error while downloading: {uri}\nHTTP Status {str(r.status_code)}", PolychromaticBase.dbg.warning, 1)
    return ""


# TODO: Refactor later. New class name
class TabData(PolychromaticBase):
    """
    This parent class is inherited by all tab objects storing common variables.
    """
    def __init__(self, appdata):
        self.appdata = appdata
        self.widgets = PolychromaticWidgets(appdata)
        self.main_window = appdata.main_window
        self.menubar = appdata.menubar
        self.header_title = appdata.main_window.findChild(QLabel, "HeaderText")

    def set_tab(self):
        """
        Called when the tab is opened via the user interface. Each deriving class
        must implement this.
        """
        raise NotImplementedError

    def set_cursor_normal(self):
        self.main_window.unsetCursor()

    def set_cursor_busy(self):
        self.main_window.setCursor(Qt.BusyCursor)

    def set_title(self, title):
        """
        Change the title on the header that's just below the menu bar.
        """
        self.header_title.setText(title)

    def create_widget_wrapper_for_control(self, widgets=[]):
        """
        Returns a widget surrounding the input list of elements, so they
        are neatly arranged for display on a row.
        """
        widget = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        widget.setLayout(layout)
        for child in widgets:
            layout.addWidget(child)
        layout.addStretch()
        return widget


# TODO: Review and move to new ControllerBase class?
class PolychromaticWidgets(PolychromaticBase):
    """
    Code for building some of the common UI elements of the Controller application.
    """
    def __init__(self, appdata):
        """
        Stores the ApplicationData() object for reference later.
        """
        self.appdata = appdata

        # Dialogues types
        self.dialog_generic = QMessageBox.Information
        self.dialog_error = QMessageBox.Critical
        self.dialog_warning = QMessageBox.Warning

    def get_icon_qt(self, folder, name):
        """
        Returns a QIcon() object with the specified button image.

        Params:
            folder          For common.get_icon()
            name            For common.get_icon()
        """
        icons = common.get_icon_styles(self.appdata.dbg, folder, name, self.appdata.normal_colour, self.appdata.disabled_colour, self.appdata.active_colour, self.appdata.selected_colour, self.appdata.secondary_colour_active, self.appdata.secondary_colour_inactive)
        if not icons:
            return QIcon()

        qicon = QIcon(icons[0])
        qicon.addFile(icons[1], mode=QIcon.Disabled)
        qicon.addFile(icons[2], mode=QIcon.Active)
        qicon.addFile(icons[2], mode=QIcon.Active, state=QIcon.On)
        qicon.addFile(icons[3], mode=QIcon.Selected)
        qicon.addFile(icons[3], mode=QIcon.Selected, state=QIcon.On)
        return qicon

    def set_toolbar_style(self, widget):
        """
        Changes a button or widget to match a toolbar preference, accepts:
        QPushButton, QToolButton, QToolBar

        For a QPushButton, "Text under icons" won't be used as it is not
        supported for this kind of button.
        """
        preferred_style = self.preferences["controller"]["toolbar_style"]

        if type(widget) in [QToolButton, QToolBar]:
            poly_to_qt = {
                pref.TOOLBAR_STYLE_DEFAULT: Qt.ToolButtonFollowStyle,
                pref.TOOLBAR_STYLE_ICONS_ONLY: Qt.ToolButtonIconOnly,
                pref.TOOLBAR_STYLE_TEXT_ONLY: Qt.ToolButtonTextOnly,
                pref.TOOLBAR_STYLE_ALONGSIDE: Qt.ToolButtonTextBesideIcon,
                pref.TOOLBAR_STYLE_UNDER: Qt.ToolButtonTextUnderIcon
            }
            widget.setToolButtonStyle(poly_to_qt[preferred_style])
            return

        if not type(widget) == QPushButton:
            raise TypeError

        if preferred_style == pref.TOOLBAR_STYLE_DEFAULT:
            # Determine what is currently in use then match a compatible equivalent.
            app_style = self.appdata.main_window.toolButtonStyle()
            try:
                qt_to_poly = {
                    Qt.ToolButtonIconOnly: pref.TOOLBAR_STYLE_ICONS_ONLY,
                    Qt.ToolButtonTextOnly: pref.TOOLBAR_STYLE_TEXT_ONLY,
                    Qt.ToolButtonTextBesideIcon: pref.TOOLBAR_STYLE_ALONGSIDE,
                }
                preferred_style = qt_to_poly[app_style]
            except KeyError:
                preferred_style = pref.TOOLBAR_STYLE_ALONGSIDE

        if preferred_style == pref.TOOLBAR_STYLE_ICONS_ONLY:
            if not widget.icon().isNull():
                widget.setToolTip(widget.text() + ' ' + widget.toolTip())
                widget.setText("")

        elif preferred_style == pref.TOOLBAR_STYLE_TEXT_ONLY:
            if len(widget.text()) > 0:
                widget.setIcon(QIcon())

    def set_dialog_buttons_icons(self, qdialogbuttonbox):
        """
        Applies Polychromatic styling to buttons inside a QDialogButtonBox.
        """
        if self.appdata.system_qt_theme:
            return

        icons = {
            QDialogButtonBox.Ok: ["general", "ok"],
            QDialogButtonBox.Cancel: ["general", "cancel"],
            QDialogButtonBox.Yes: ["general", "ok"],
            QDialogButtonBox.No: ["general", "cancel"],
            QDialogButtonBox.Retry: ["general", "refresh"],
            QDialogButtonBox.Ignore: ["general", "cancel"],
            QDialogButtonBox.Save: ["general", "save"],
            QDialogButtonBox.Discard: ["general", "delete"],
            QDialogButtonBox.RestoreDefaults: ["general", "reset"],
        }

        for std_btn in icons.keys():
            button = qdialogbuttonbox.button(std_btn)
            if button:
                try:
                    button.setIcon(self.get_icon_qt(icons[std_btn][0], icons[std_btn][1]))
                except KeyError:
                    # None specified. Strip icon (to avoid native theme mixing)
                    button.setIcon(QIcon())

    def create_summary_widget(self, icon_path, title, indicators=[], buttons=[]):
        """
        Returns a summary widget presenting an overview of the selected device,
        effect or preset.

        Params:
            icon_path           (str)   Absolute path to the icon to use.
            icon_path_fallback  (str)   Absolute path when icon cannot be found.
            title               (str)   Summary title, e.g. device or effect name
            indicators          (list)  See indicators below.
            buttons             (list)  See buttons below.

        Indicators list:
        [
            {
                "icon": (str)       (Absolute path)
                "label": (str)
            },
            {..}
        ]

        Buttons list:
        [
            {
                "id": (str)
                "icon": (QIcon)
                "label": (str)
                "disabled": (bool)
                "action": (obj)     (Function to run when clicked)
            },
            {..}
        ]

        Returns: QWidget()
        """
        summary = get_ui_widget(self.appdata, "widget-summary")
        container_widget = summary.findChild(QWidget, "SummaryContainer")
        image_widget = summary.findChild(QLabel, "SummaryImage")
        title_widget = summary.findChild(QLabel, "SummaryTitle")
        indicators_widget = summary.findChild(QWidget, "SummaryIcons")
        buttons_widget = summary.findChild(QWidget, "SummaryButtons")

        title_widget.setText(title)
        title_widget.setStyleSheet("QLabel {{ font-size: 17px; {0} }}".format("color: lime;" if not self.appdata.system_qt_theme else ""))

        if self.appdata.system_qt_theme:
            summary.setStyleSheet("#Summary {{ background: {0} }}".format(summary.palette().alternateBase().color().name()))

        # Populate Image
        if os.path.exists(icon_path):
            if icon_path.endswith(".svg"):
                # Replace summary icon with SVG renderer for better quality/scaling
                if icon_path.endswith(".svg"):
                    summary.findChild(QLabel, "SummaryImage").deleteLater()
                    viewer = QSvgWidget()
                    viewer.load(icon_path)
                    viewer.setMinimumHeight(90)
                    viewer.setMinimumWidth(90)
                    viewer.setMaximumHeight(90)
                    viewer.setMaximumWidth(90)
                    summary.setContentsMargins(8 + 6, 8, 8, 8)

                    # Appends to the end - restructure
                    summary.layout().addWidget(viewer)
                    summary.layout().addWidget(container_widget)
                    summary.layout().addWidget(buttons_widget)
                    viewer.show()
            else:
                set_pixmap_for_label(image_widget, icon_path, 115)
        else:
            image_widget.deleteLater()

        # Populate Indicators
        for indicator in indicators:
            widget = QWidget()
            widget.setLayout(QHBoxLayout())
            layout = widget.layout()
            layout.setContentsMargins(0, 0, 10, 0)

            # Create image
            if indicator["icon"]:
                label = QLabel()
                set_pixmap_for_label(label, indicator["icon"], 22)
                layout.addWidget(label)

            # Create text
            label = QLabel()
            label.setText(indicator["label"])
            layout.addWidget(label)

            indicators_widget.layout().addWidget(widget)
        indicators_widget.layout().addStretch()

        # Populate Buttons
        for btn in buttons:
            button = QPushButton()
            button.setObjectName(btn["id"])
            if btn["icon"]:
                button.setIcon(btn["icon"])
            button.setText(btn["label"])

            if btn["disabled"]:
                button.setDisabled(True)

            button.clicked.connect(btn["action"])
            buttons_widget.layout().addWidget(button)
        buttons_widget.layout().addStretch()

        return summary

    def create_group_widget(self, title):
        """
        Returns a group widget containing controls for a specific device zone.

        Params:
            title           (str)   Group title

        Returns: QGroupBox()
        """
        group = QGroupBox()
        group.setTitle(title)
        group.setLayout(QGridLayout())
        group.layout().setAlignment(Qt.AlignTop)
        group.layout().setContentsMargins(QMargins(0, 6, 0, 6))
        return group

    def create_row_widget(self, label_text, widgets=[], vertical=False, wrap=False):
        """
        Returns a widget for use when presenting controls for a particular option.

        Params:
            label_text      (str)   Human readable control label
            widgets         (list)  List of widgets to display
            vertical        (bool)  Use a vertical layout instead of horizontal
            wrap            (bool)  Wrap widgets when they get too long (horizontal only)

        Returns: QWidget()
        """
        widget = QWidget()
        widget.setLayout(QHBoxLayout() if wrap else QFormLayout())
        layout = widget.layout()
        layout.setContentsMargins(QMargins(30, 5, 30, 5))

        # Left - Add label
        label = QLabel()
        label.setText(label_text)
        label.setAlignment(Qt.AlignTop)
        label.setWordWrap(True)
        label.setMinimumWidth(120)
        label.setMaximumWidth(120)

        # Right - Add controls
        inner_widget = QWidget()
        if wrap:
            inner_widget.setLayout(QFlowLayout())
        else:
            inner_widget.setLayout(QVBoxLayout() if vertical else QHBoxLayout())

        inner_widget.layout().setContentsMargins(0, 0, 0, 0)
        for w in widgets:
            inner_widget.layout().addWidget(w)

        if wrap:
            layout.addWidget(label)
            layout.addWidget(inner_widget)
        else:
            layout.addRow(label, inner_widget)

        return widget

    def populate_empty_state(self, layout, icon_path, title, subtitle, buttons=[]):
        """
        Creates a 'watermark' like background when a page or section is empty.
        For example, this may be used when the device list is empty.

        Params:
            layout          QWidget     QWidget() to apply the layout on. Should be empty.
            icon_path       (str)       Path to the background graphic. Should be 750x500.
            title           (str)       Larger text to display.
            subtitle        (str)       Smaller text to display.
            buttons         (list)      (Optional) In format: [{icon_folder, icon_name, label, action}, {..}]

        Returns nothing - the layout is directly applied.
        """
        # Center graphic
        image = QLabel()
        pixmap_src = QPixmap(icon_path)
        pixmap = pixmap_src.scaled(int(750 / 2), int(500 / 2), Qt.KeepAspectRatio, Qt.SmoothTransformation)
        image.setPixmap(pixmap)
        image.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        image.setAlignment(Qt.AlignCenter)

        # Text & buttons
        text1 = QLabel()
        text1.setFont(QFont("Play", 14))
        text1.setAlignment(Qt.AlignCenter)
        text1.setText(title)
        text1.setMargin(10)

        text2 = QLabel()
        text2.setFont(QFont("Play", 11))
        text2.setAlignment(Qt.AlignCenter)
        text2.setText(subtitle)
        text2.setMargin(10)
        text2.setWordWrap(True)

        button_container = None
        if buttons:
            button_container = QWidget()
            button_container.setLayout(QHBoxLayout())
            button_container.layout().addStretch()
            for spec in buttons:
                button = QPushButton()
                button.setText(spec["label"])
                button.setIcon(self.get_icon_qt(spec["icon_folder"], spec["icon_name"]))
                button.clicked.connect(spec["action"])
                button_container.layout().addWidget(button)
            button_container.layout().addStretch()

        # Putting it together
        layout.addStretch()
        for widget in [image, text1, text2, button_container]:
            if widget:
                layout.addWidget(widget)
        layout.addStretch()

    def create_colour_control(self, current_hex, callback_fn, callback_data, title, monoscale=False):
        """
        Create a colour picker control for the user to set a colour. Setting
        the colour will open a dialog.

        When a colour is saved, the callback_fn will be executed passing the new
        hex colour as a parameter as well as callback_data (as one variable)

        Params:
            current_hex     String of the current hex value in use.
            callback_fn     Function to run after saving changes.
            callback_data   Additional data to pass to callback_fn.
            title           Title to show when colour is being picked.
            monoscale       Picker should only show green only colours.
        """
        container = QWidget()
        container.setLayout(QHBoxLayout())
        container.layout().setContentsMargins(0, 0, 0, 0)
        preview = QWidget(objectName="PickerPreview")
        preview.setMinimumHeight(28)
        preview.setMaximumHeight(28)
        preview.setMinimumWidth(80)
        preview.setMaximumWidth(80)
        preview.setStyleSheet("QWidget {{ background-color: {0} }}".format(current_hex))

        btn = QPushButton(objectName="PickerCustom")
        btn.setText(self.appdata._("Change..."))
        btn.current_hex = current_hex

        # Dynamic function allows changing the initial colour upon opening the picker later.
        def _change_colour(new_hex, data=None):
            btn.current_hex = new_hex
            preview.setStyleSheet("QWidget {{ background-color: {0} }}".format(new_hex))
            callback_fn(new_hex, data)
        btn.change_colour = _change_colour

        # Connect the signal
        def _clicked_change_colour():
            picker = ColourPicker(self.appdata, _change_colour, callback_data, btn.current_hex, title, monoscale, preview)
        btn.clicked.connect(_clicked_change_colour)

        # Put it all together
        container.layout().addWidget(preview)
        container.layout().addWidget(btn)
        container.layout().addStretch()
        return container

    def create_icon_picker_control(self, callback_fn, current_icon, title, purpose=0):
        """
        Create an icon picker control for the user to choose an icon from a list
        of built-in icons, user-imported icons ("custom icons") or an application
        installed on the file system.

        When an icon is saved, the callback_fn will be executed returning the
        relative path (for a built-in icon) or an absolute path (for custom images)
        as a parameter for the function that will process the new selection.

        Params:
            callback_fn     Function to run after saving changes
            current_icon    Initial icon
            title           Window title
            purpose         IconPicker.purpose_* integer.
        """
        container = QWidget()
        container.setLayout(QHBoxLayout())
        container.layout().setContentsMargins(0, 0, 0, 0)

        preview = QLabel()
        preview.setMaximumHeight(24)
        preview.setMaximumWidth(24)
        preview.current_icon = current_icon
        set_pixmap_for_label(preview, common.get_full_path_for_save_data_icon(preview.current_icon), 24)

        # TODO: Improve icon quality on HiDPI displays

        def _changed_icon_callback(new_icon):
            set_pixmap_for_label(preview, common.get_full_path_for_save_data_icon(new_icon), 24)
            callback_fn(new_icon)
            preview.current_icon = new_icon

        def _clicked_change_icon():
            picker = IconPicker(self.appdata, _changed_icon_callback, preview.current_icon, title, purpose)

        btn = QPushButton()
        btn.setText(self.appdata._("Change..."))
        btn.clicked.connect(_clicked_change_icon)

        container.layout().addWidget(preview)
        container.layout().addWidget(btn)
        container.layout().addStretch()
        return container

    def open_dialog(self, dialog_type, title, text, info_text="", details="", buttons=[QMessageBox.Ok], default_button=None, actions={}):
        """
        Opens a modal dialogue box to inform or confirm a decision.

        If the buttons are unspecified, the dialog will just show a message
        with an "OK" button that takes no further action.

        Params:
            dialog_type     (str)   One of self.dialog_* variable above.
            title           (str)   Window title
            text            (str)   Description of what happened.
            info_text       (str)   (Optional) Informative text of next steps.
                                    Appears under text. macOS styles differently.
            details         (str)   (Optional) Monospace output of the problem
            buttons         (list)  (Optional) QMessageBox.* of buttons to display
            default_button  (obj)   (Optional) QMessageBox.* of initial default button
            actions         (dict)  (Optional) Bind a QMessageBox.* to function
        """
        msgbox = QMessageBox()
        msgbox.setWindowTitle(title)
        msgbox.setWindowFlag(Qt.Popup)
        msgbox.setText(text)
        msgbox.setIcon(dialog_type)

        if not self.appdata.system_qt_theme:
            msgbox.setPalette(get_palette(self.appdata))
            load_qt_theme(self.appdata, msgbox)

        if info_text:
            msgbox.setInformativeText(info_text);

        for std_btn in buttons:
            msgbox.addButton(std_btn)
            button = msgbox.buttons()[-1]

            if std_btn == default_button:
                msgbox.setDefaultButton(button)

            if std_btn in actions.keys():
                button.clicked.connect(actions[std_btn])

        if details:
            msgbox.setDetailedText(details)
            msgbox.findChild(QTextEdit).setMinimumWidth(600)

        if not self.appdata.system_qt_theme:
            icon_path = common.get_icon("dialog", "info")
            if dialog_type == self.dialog_error:
                icon_path = common.get_icon("dialog", "error")
            elif dialog_type == self.dialog_warning:
                icon_path = common.get_icon("dialog", "warning")

            msgbox.setIconPixmap(QPixmap(icon_path))

        if self.appdata.system_qt_theme:
            msgbox.setStyleSheet("QTextEdit { font-family: monospace; }")

        dialog_buttons = msgbox.findChild(QDialogButtonBox)
        self.set_dialog_buttons_icons(dialog_buttons)

        msgbox.exec()


class ColourPicker(PolychromaticBase):
    """
    The colour picker dialog allows the user to quickly choose a colour or
    hand over to the system's colour picker (which on Linux, would be Qt's native picker)
    """
    def __init__(self, appdata, callback_fn, callback_data, current_hex, title, monoscale, preview):
        """
        Params:
            appdata         (obj)   ApplicationData() object
            callback_fn     (obj)   Run this function after saving new colour. The parameters are (new_hex, callback_data)
            callback_data   (any)   Additional data to pass to callback_fn
            current_hex     (str)   Starting colour value (#RRGGBB)
            title           (str)   Window title
            monoscale       (bool)  Only show green shades (for monochromatic devices)
            preview         (obj)   QWidget() of the preview box
        """
        self.appdata = appdata
        self.widgets = PolychromaticWidgets(appdata)
        self.current_hex = current_hex
        self.current_name = ""
        self.callback_fn = callback_fn
        self.callback_data = callback_data
        self.title = title
        self.monoscale = monoscale
        self.preview = preview
        self.add_mode = "append"

        # UI Controls
        self.dialog = get_ui_widget(appdata, "colour-picker", q_toplevel=QDialog)
        self.dialog_btns = self.dialog.findChild(QDialogButtonBox, "buttonBox")
        self.change_btn = self.dialog.findChild(QPushButton, "OpenPicker")
        self.list_save_btn = self.dialog.findChild(QPushButton, "SaveToList")
        self.list_del_btn = self.dialog.findChild(QPushButton, "DeleteFromList")
        self.list_rename_btn = self.dialog.findChild(QPushButton, "RenameFromList")
        self.list_text_input = self.dialog.findChild(QLineEdit, "ColourName")
        self.save_widget = self.dialog.findChild(QWidget, "SaveWidget")
        self.save_widget.setHidden(True)
        self.open_save_widget = self.dialog.findChild(QPushButton, "OpenSaveWidget")
        self.close_save_widget = self.dialog.findChild(QPushButton, "CloseSaveWidget")
        self.saved_tree = self.dialog.findChild(QTreeWidget, "SavedColours")
        self.current_preview = self.dialog.findChild(QWidget, "CurrentPreview")
        self.current_label = self.dialog.findChild(QLabel, "CurrentLabel")

        # Set Dialog Button Icons
        if not self.appdata.system_qt_theme:
            self.change_btn.setIcon(self.widgets.get_icon_qt("general", "edit"))
            self.open_save_widget.setIcon(self.widgets.get_icon_qt("general", "new"))
            self.close_save_widget.setIcon(self.widgets.get_icon_qt("general", "close"))
            self.list_save_btn.setIcon(self.widgets.get_icon_qt("general", "save"))
            self.list_del_btn.setIcon(self.widgets.get_icon_qt("general", "delete"))
            self.list_rename_btn.setIcon(self.widgets.get_icon_qt("general", "properties"))

        # If the device only supports a monoscale of "RGB", use a fixed list.
        if monoscale:
            self.list_del_btn.setDisabled(True)
            self.list_rename_btn.setDisabled(True)
            self.open_save_widget.setDisabled(True)
            self.open_save_widget.setHidden(True)
            self.dialog.findChild(QWidget, "SavedColoursVList").setHidden(True)

        # Connect signals when interacting with UI controls
        self.dialog_btns.accepted.connect(self._apply_colour)
        self.change_btn.clicked.connect(self._open_system_picker)
        self.open_save_widget.clicked.connect(self._open_save_widget)
        self.close_save_widget.clicked.connect(self._close_save_widget)
        self.list_text_input.textEdited.connect(self._on_save_input_change)
        self.list_text_input.returnPressed.connect(self._save_to_list)
        self.list_save_btn.clicked.connect(self._save_to_list)
        self.list_del_btn.clicked.connect(self._delete_from_list)
        self.list_rename_btn.clicked.connect(self._rename_list_item)
        self.saved_tree.itemSelectionChanged.connect(self._switch_colour)

        # Refresh UI data and open dialog
        self._build_saved_colour_list()
        self._refresh_selected_colour(current_hex)
        self.saved_tree.setColumnWidth(0, 140)
        self.dialog.setWindowTitle(title)
        self._select_current_colour()
        self.dialog.exec()

    def _switch_colour(self):
        """
        User selects a colour from their selected colour list.
        """
        items = self.saved_tree.selectedItems()
        if items:
            self._refresh_selected_colour(items[0].colour_hex)

    def _add_to_tree(self, name, value):
        """
        Appends an item to the tree.
        """
        item = QTreeWidgetItem()
        item.setText(0, name)
        item.setText(1, value.upper())
        item.setIcon(0, QIcon(common.generate_colour_bitmap(self.appdata.dbg, value, 20)))
        item.colour_name = name
        item.colour_hex = value.upper()
        item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsDragEnabled)
        self.saved_tree.addTopLevelItem(item)
        return item

    def _get_tree_objects(self):
        root = self.saved_tree.invisibleRootItem()
        items = []
        for i in range(root.childCount()):
            items.append(root.child(i))
        return items

    def _refresh_selected_colour(self, new_hex):
        """
        Refresh the UI to show the newly chosen colour.
        """
        self.current_hex = new_hex
        self.current_preview.setStyleSheet("QWidget {{ background-color: {0} }}".format(new_hex))
        self.current_label.setText("{0}\n({1})".format(new_hex.upper(), ", ".join(map(str, common.hex_to_rgb(new_hex)))))

        self.list_text_input.setText("")
        self.open_save_widget.setDisabled(False)
        self.list_del_btn.setDisabled(True)
        self.list_rename_btn.setDisabled(True)

        for widget in self._get_tree_objects():
            if widget.colour_hex == new_hex:
                widget.setSelected(True)
                self.current_name = widget.colour_name
                self.save_widget.setHidden(True)
                self.open_save_widget.setDisabled(True)
                self.list_del_btn.setDisabled(False)
                self.list_rename_btn.setDisabled(False)
                return

        self.saved_tree.clearSelection()

    def _apply_colour(self):
        """
        User applies their newly chosen colour. Saves the colour list to file if
        there are new colours or the order was modified.
        """
        dbg = self.appdata.dbg
        self._save_colour_list_to_file()
        self.preview.setStyleSheet("QWidget {{ background-color: {0} }}".format(self.current_hex))
        dbg.stdout("Colour set to: " + self.current_hex, dbg.success, 1)
        self.callback_fn(self.current_hex, self.callback_data)

    def _save_colour_list_to_file(self):
        """
        Saves the colour list to file.
        """
        # When pre-defined monochromatic colours are used, never save!
        if self.monoscale:
            return

        # TODO: Optimisation: Only save if there are changes (add/remove/reorder)
        dbg = self.appdata.dbg
        dbg.stdout("Saving colour list...", dbg.action, 1)
        colour_list = []
        for item in self._get_tree_objects():
            colour_list.append({
                "name": item.colour_name,
                "hex": item.colour_hex
            })
        pref.save_file(self.paths.colours, colour_list)

        # Reload the tray applet to use new colour list
        process = procpid.ProcessManager("tray-applet")
        if process.is_another_instance_is_running():
            dbg.stdout("Colour list saved. Reloading tray applet...", dbg.action, 1)
            process.reload()

    def _open_system_picker(self):
        """
        User uses the system's colour picker to pick any colour, which could
        include the ability to choose a pixel on the screen (dependent on OS)
        """
        dialog = QColorDialog()

        # TODO: Nice to have: Use Polychromatic styling for the dialog
        # QColorDialog() is private, so this may be tricky.
        # This does not work!
        load_qt_theme(self.appdata, dialog)

        output = dialog.getColor(title=self.title, initial=QColor(self.current_hex))

        if not output.isValid():
            return
        new_hex = output.name()

        # Strip R/B out of RGB if device is monochromatic
        if self.monoscale:
            as_rgb = common.hex_to_rgb(new_hex)
            new_hex = common.rgb_to_hex([0, as_rgb[1], 0])

        self._refresh_selected_colour(new_hex)

    def _open_save_widget(self, replace=False):
        self.save_widget.setHidden(False)
        self.open_save_widget.setDisabled(True)

        self.add_mode = "append"
        if replace:
            self.add_mode = "replace"

    def _close_save_widget(self):
        self.save_widget.setHidden(True)
        self.open_save_widget.setDisabled(self.add_mode == "replace")

    def _on_save_input_change(self, text):
        self.list_save_btn.setEnabled(True if len(text) > 0 else False)

    def _save_to_list(self):
        """
        Save the colour to their Saved Colour list.
        """
        new_name = self.list_text_input.text()
        new_hex = self.current_hex

        if self.add_mode == "append":
            item = self._add_to_tree(new_name, new_hex)
            item.setSelected(True)
        elif self.add_mode == "replace":
            item = self.saved_tree.selectedItems()[0]
            item.setText(0, new_name)

        self._close_save_widget()

    def _delete_from_list(self):
        """
        Delete the selected item from the Saved Colour list.
        """
        try:
            item = self.saved_tree.selectedItems()[0]
            self.saved_tree.invisibleRootItem().removeChild(item)
        except IndexError:
            # UI Hiccup - clicked too fast, last item in list, etc
            self.list_del_btn.setDisabled(True)
            self.list_rename_btn.setDisabled(True)

    def _rename_list_item(self):
        """
        Rename the selected colour from the Saved Colour list.
        """
        try:
            item = self.saved_tree.selectedItems()[0]
            name = item.text(0)
            self.list_text_input.setText(name)
            self._open_save_widget(replace=True)
        except IndexError:
            # UI Hiccup - clicked too fast, last item in list, etc
            self.list_del_btn.setDisabled(True)
            self.list_rename_btn.setDisabled(True)

    def _build_saved_colour_list(self):
        """
        Validates the colours and builds the initial saved colour list.
        """
        self.saved_tree.invisibleRootItem().takeChildren()

        if self.monoscale:
            colour_list = common.get_green_shades(self._)
        else:
            colour_list = pref.get_colour_list(self._)

        for index, colour in enumerate(colour_list):
            item = self._add_to_tree(colour["name"], colour["hex"])

    def _select_current_colour(self):
        """
        Once Qt initialises, we're ready to select the current colour from the list.
        """
        tree = self.saved_tree.invisibleRootItem()
        for index, item in enumerate(tree.takeChildren()):
            tree.addChild(item)
            if item.colour_hex.upper() == self.current_hex.upper():
                item.setSelected(True)
                if index > 5:
                    self.saved_tree.scrollToItem(item)


class IconPicker(PolychromaticBase):
    """
    The icon picker dialog allows the user to choose an icon for an effect,
    preset or tray applet.

    For convenience, installed applications and Steam games will be listed as
    icon sources (excluded in tray applet)
    """
    PURPOSE_GENERIC = 0
    PURPOSE_TRAY_ONLY = 1

    INDEX_TRAY = 0
    INDEX_EMBLEMS = 1
    INDEX_APPS = 2
    INDEX_STEAM = 3
    INDEX_CUSTOM = 4

    def __init__(self, appdata, callback_fn, current_icon, title, purpose=0):
        """
        Params:
            appdata         (obj)   ApplicationData() object
            callback_fn     (obj)   Run this function after saving new colour. The result will be passed as a parameter.
            current_icon    (str)   Initial icon path (usually relative)
            title           (str)   Window title
            purpose         (bool)  One of the IconPicker.* integers to specify intended behaviour.
        """
        self.appdata = appdata
        self.widgets = PolychromaticWidgets(appdata)
        self.current_icon = current_icon
        self.callback_fn = callback_fn
        self.title = title
        self.purpose = purpose
        self.custom_icon_exts = ["svg", "png", "jpg", "jpeg", "gif"]
        self.custom_icon_filters = [
            self._("Image Files") + "(*.png *.jpg *.jpeg *.gif *.svg)",
            self._("PNG Image") + " (*.png)",
            self._("JPEG Image") + " (*.jpg *.jpeg)",
            self._("GIF Image") + " (*.gif)",
            self._("SVG Image") + " (*.svg)",
            self._("All Files") + " (*)"
        ]
        self.custom_empty = True

        # UI Controls
        self.dialog = get_ui_widget(appdata, "icon-picker", q_toplevel=QDialog)
        self.dialog_btns = self.dialog.findChild(QDialogButtonBox, "buttonBox")
        self.tabs = self.dialog.findChild(QTabWidget, "IconTabs")
        self.custom_icon_add = self.dialog.findChild(QPushButton, "ImportCustomIcon")
        self.custom_icon_del = self.dialog.findChild(QPushButton, "DeleteCustomIcon")
        self.gif_warning = self.dialog.findChild(QWidget, "AnimatedGIFWarningLabel")
        self.button_group = QButtonGroup()
        self.gif_previews = []

        # Tab contents for housing the icons
        self.icons_tray = self.dialog.findChild(QWidget, "TraySet")
        self.icons_emblems = self.dialog.findChild(QWidget, "EmblemSet")
        self.icons_apps = self.dialog.findChild(QWidget, "ApplicationSet")
        self.icons_steam = self.dialog.findChild(QWidget, "SteamSet")
        self.icons_custom = self.dialog.findChild(QWidget, "CustomSet")
        self.tab_index_widgets = {
            self.INDEX_TRAY: self.icons_tray,
            self.INDEX_EMBLEMS: self.icons_emblems,
            self.INDEX_APPS: self.icons_apps,
            self.INDEX_STEAM: self.icons_steam,
            self.INDEX_CUSTOM: self.icons_custom
        }

        # Set Dialog Button Icons
        if not self.appdata.system_qt_theme:
            self.custom_icon_add.setIcon(self.widgets.get_icon_qt("general", "folder"))
            self.custom_icon_del.setIcon(self.widgets.get_icon_qt("general", "delete"))
            self.tabs.setTabIcon(self.INDEX_TRAY, self.widgets.get_icon_qt("general", "tray-applet"))
            self.tabs.setTabIcon(self.INDEX_EMBLEMS, self.widgets.get_icon_qt("emblems", "misc"))
            self.tabs.setTabIcon(self.INDEX_APPS, self.widgets.get_icon_qt("emblems", "software"))
            self.tabs.setTabIcon(self.INDEX_STEAM, self.widgets.get_icon_qt("emblems", "steam"))
            self.tabs.setTabIcon(self.INDEX_CUSTOM, self.widgets.get_icon_qt("general", "folder"))

        # Gather icon data
        list_tray = pref.load_file(os.path.join(self.paths.data_dir, "img", "tray", "icons.json"))
        list_emblems = pref.load_file(os.path.join(self.paths.data_dir, "img", "emblems", "icons.json"))
        if not self.purpose == self.PURPOSE_TRAY_ONLY:
            list_apps = self._get_application_icons()
            list_steam = self._get_steam_icons()
        list_custom = glob.glob(self.paths.custom_icons + "/*")
        self.custom_empty = True if len(list_custom) == 0 else False

        # Populate icon tabs
        all_icon_buttons = []
        self._load_icon_set(self.INDEX_EMBLEMS, list_emblems)
        self._load_icon_set(self.INDEX_CUSTOM, list_custom)
        if self.purpose == self.PURPOSE_TRAY_ONLY:
            self._load_icon_set(self.INDEX_TRAY, list_tray)
        else:
            self._load_icon_set(self.INDEX_APPS, list_apps)
            self._load_icon_set(self.INDEX_STEAM, list_steam)

        # When changing tray applet icon, limit the selection
        if self.purpose == self.PURPOSE_TRAY_ONLY:
            self.tabs.removeTab(self.INDEX_STEAM)
            self.tabs.removeTab(self.INDEX_APPS)
        else:
            self.tabs.removeTab(self.INDEX_TRAY)

        # TODO: Scroll to selected item

        # Prepare and open dialog
        self._setup_drag_drop_custom_icon()
        self.gif_warning.setHidden(True)
        self.button_group.buttonClicked.connect(self.select_icon)
        self.custom_icon_add.clicked.connect(self.import_custom_icon)
        self.custom_icon_del.clicked.connect(self.delete_custom_icon)

        self.dialog.accepted.connect(self.accept_changes)
        self.dialog.finished.connect(self.close_dialog)
        self.dialog.setWindowTitle(title)
        self.dialog.exec()

    def close_dialog(self):
        """
        Ensure the dialog (and its children) are destroyed. This is because the
        tray applet has some animated icons that have QMovie children.
        """
        for movie in self.gif_previews:
            movie.stop()
            movie.deleteLater()
        self.dialog.deleteLater()

    def _load_icon_set(self, tab_index, icon_list):
        """
        Populates the icons for the specified tab.

        Each button will be injected a variable for save data purposes:
            .icon_path      Relative/absolute path
            .tab_index      Tab index (for setting active state)
        """
        # No icons for category?
        if len(icon_list) == 0:
            return self._load_empty_set(tab_index)

        tab_widget = self.tab_index_widgets[tab_index]
        tab_widget.setLayout(QFlowLayout())

        # Tray applet supports animated GIFs.
        def _load_animated_gif(button):
            movie = QMovie(common.get_full_path_for_save_data_icon(icon_path))
            movie.button = button

            def _draw_gif(frame):
                movie.button.setIcon(QIcon(movie.currentPixmap()))

            def _loop_gif(movie):
                movie.start()

            self.gif_previews.append(movie)
            movie.frameChanged.connect(_draw_gif)
            movie.finished.connect(_loop_gif)
            movie.start()

        # Populate the list
        for icon_path in icon_list:
            button = self._make_icon_button(icon_path, tab_index)
            tab_widget.layout().addWidget(button)

            if icon_path.endswith(".gif"):
                _load_animated_gif(button)

        # Set initial selection
        for button in self.button_group.buttons():
            if button.icon_path == self.current_icon:
                button.setChecked(True)
                self.tabs.setCurrentIndex(button.tab_index)

                if tab_index == self.INDEX_CUSTOM:
                    self.custom_icon_del.setEnabled(True)

                # TODO: Scroll down to item. Refactor into list widget?

                break

    def _load_empty_set(self, tab_index):
        """
        Shows an empty message for the specified tab when there are no icons
        available to choose from.
        """
        tab_widget = self.tab_index_widgets[tab_index]
        tab_widget.setLayout(QVBoxLayout())

        widget = QWidget()
        widget.setObjectName("empty_" + str(tab_index))
        widget.setLayout(QVBoxLayout())
        widget.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)

        label = QLabel()
        label.setWordWrap(True)
        label.setAlignment(Qt.AlignHCenter)
        label.setMargin(4)

        images = {
            self.INDEX_STEAM: common.get_icon("empty", "steam"),
            self.INDEX_CUSTOM: common.get_icon("empty", "custom"),
        }

        messages = {
            self.INDEX_STEAM: self._("When Steam is installed, icons from your games will appear here."),
            self.INDEX_CUSTOM: self._("Drag and drop icons here, or add them by pressing the Import button.")
        }

        try:
            label.setText(messages[tab_index])
        except KeyError:
            label.setText(self._("No icons found!"))

        image = QLabel()
        image.setAlignment(Qt.AlignHCenter)
        if tab_index in images.keys():
            image.setPixmap(QPixmap(images[tab_index]))

        widget.layout().addStretch()
        widget.layout().addWidget(image)
        widget.layout().addWidget(label)
        widget.layout().addStretch()
        return tab_widget.layout().addWidget(widget)

    def _make_icon_button(self, icon_path, tab_index):
        """
        Creates an icon button for selection.
        """
        button = QToolButton()
        button.setToolTip(icon_path.split("/")[-1])
        button.setCheckable(True)
        button.setIconSize(QSize(32, 32))
        button.setIcon(QIcon(common.get_full_path_for_save_data_icon(icon_path)))
        button.icon_path = icon_path
        button.tab_index = tab_index
        self.button_group.addButton(button)

        return button

    def select_icon(self, button):
        """
        Updates the current icon choice in memory. When a GIF (tray applet) is selected,
        inform the user of the possibility it might not work on their desktop environment.
        """
        self.current_icon = button.icon_path
        gif_selected = button.tab_index == 0 and button.icon_path.endswith(".gif")
        self.gif_warning.setHidden(not gif_selected)

        # Only custom icons can be deleted
        self.custom_icon_del.setEnabled(button.tab_index == 4)

    def accept_changes(self):
        """
        Accepts the new icon choice and closes the icon picker. Runs the callback
        function to process the result.
        """
        self.callback_fn(self.current_icon)
        self.close_dialog()

    def _setup_drag_drop_custom_icon(self):
        """
        Prepare the file drag and drop facility as an alternate way to import
        custom icons.
        """
        def dragEnterEvent(event):
            if event.mimeData().hasUrls():
                event.accept()
            else:
                event.ignore()

        def dropEvent(event):
            uris = event.mimeData().urls()
            for uri in uris:
                if uri.isLocalFile() and uri.path().split(".")[-1].lower() in self.custom_icon_exts:
                    self.process_new_custom_icon(uri.path())

        self.tabs.setAcceptDrops(True)
        self.tabs.dragEnterEvent = dragEnterEvent
        self.tabs.dropEvent = dropEvent

    def process_new_custom_icon(self, source_path):
        """
        Adds the specified icon to the user's custom icons set. This may also
        be triggered by dropping graphics.
        """
        target_path = os.path.join(self.paths.custom_icons, os.path.basename(source_path))

        # Reset layout if new icons are to be added
        if self.custom_empty:
            self.custom_empty = False
            self.icons_custom.findChild(QWidget, "empty_" + str(self.INDEX_CUSTOM)).deleteLater()
            QWidget().setLayout(self.icons_custom.layout())
            self.icons_custom.setLayout(QFlowLayout())

        # Prevent duplicates
        while os.path.exists(target_path):
            self.dbg.stdout("Custom icon filename already exists: " + target_path, self.dbg.warning, 1)
            extension = os.path.basename(target_path).split(".")[-1]
            target_path = target_path.replace("." + extension, "_." + extension)

        if os.path.exists(source_path):
            self.dbg.stdout("Adding custom icon: " + source_path, self.dbg.action, 1)
            shutil.copy(source_path, target_path)
            button = self._make_icon_button(target_path, 4)
            self.tab_index_widgets[self.INDEX_CUSTOM].layout().addWidget(button)

        # Set as selected
        self.tabs.setCurrentIndex(self.tabs.count() - 1)
        button.click()

    def import_custom_icon(self):
        """
        Opens the file browser to import new files to the custom icons set.
        """
        browser = QFileDialog()
        browser.setAcceptMode(QFileDialog.AcceptOpen)
        browser.setFileMode(QFileDialog.ExistingFiles)

        # TODO: Use Polychromatic styles, if possible.
        load_qt_theme(self.appdata, browser)

        files = browser.getOpenFileNames(caption=self._("Import Custom Icon"), filter=";;".join(self.custom_icon_filters))[0]

        if len(files) == 0:
            return

        for f in files:
            if f.startswith("/"):
                self.process_new_custom_icon(f)

    def delete_custom_icon(self, b):
        """
        Removes the specified icon from the user's custom icon set.
        Any save data referencing this icon will show a generic icon instead.
        """
        button = self.button_group.checkedButton()
        if not button:
            return

        icon_path = button.icon_path
        self.dbg.stdout("Deleting custom icon: " + icon_path, self.dbg.action, 1)
        os.remove(icon_path)
        button.deleteLater()

        self.custom_icon_del.setEnabled(False)

    def _get_application_icons(self):
        """
        Parses the desktop launchers for applications both local and system-wide
        so the user can pick them.
        """
        self.dbg.stdout("Populating application icons...", self.dbg.action, 1)
        local_apps = glob.glob(os.path.expanduser("~") + "/.local/share/applications/*.desktop")
        system_apps = glob.glob("/usr/share/applications/*.desktop")
        file_list = local_apps + system_apps
        icon_list = []

        # Determine where to find application icons (from theme)
        theme_name = QIcon.themeName()
        theme_paths = QIcon.themeSearchPaths()
        search_paths = []
        already_found = []

        search_paths.append(os.path.join(os.path.expanduser("~"), ".local", "share", "icons", "**"))
        for theme_path in theme_paths:
            search_paths.append(os.path.join(theme_path, theme_name, "apps", "48"))
        search_paths.append("/usr/share/icons/hicolor/48x48/apps")
        search_paths.append("/usr/share/icons/hicolor/64x64/apps")
        search_paths.append("/usr/share/icons/hicolor/128x128/apps")
        search_paths.append("/usr/share/icons/hicolor/scalable/apps")
        search_paths.append("/usr/share/pixmaps/")

        def _find_theme_icon(icon_name):
            # Prevent duplicates
            if icon_name in already_found:
                return None

            for search_path in search_paths:
                for path in glob.glob(os.path.join(search_path, "*"), recursive=True):
                    filename = os.path.splitext(os.path.basename(path))[0]
                    if filename == icon_name and os.path.isfile(path):
                        already_found.append(icon_name)
                        return path
            return None

        def parse_launcher_for_icon(lines):
            for line in lines:
                if line.startswith("Icon="):
                    icon = line.split("Icon=")[1].strip()

                    # Absolute path
                    if os.path.exists(icon):
                        icon_list.append(icon)
                        return

                    # Search theme directories
                    theme_icon = _find_theme_icon(icon)
                    if theme_icon:
                        icon_list.append(theme_icon)

        for launcher_path in file_list:
            with open(launcher_path, "r") as f:
                parse_launcher_for_icon(f.readlines())

        self.dbg.stdout("Loaded {0} application icons from {1} launchers.".format(len(icon_list), len(file_list)), self.dbg.success, 1)

        # Sort applications A-Z
        return sorted(icon_list, key=lambda i: os.path.basename(i).lower())

    def _get_steam_icons(self):
        """
        If Steam is installed (usually ~/.steam), the application will try to
        locate "librarycache" which contains many *_icon.jpg files of the games
        the user has installed/cached.
        """
        icon_list = []
        steam_dir = os.path.expanduser("~") + "/.steam"
        if not os.path.exists(steam_dir):
            return []

        for folder in ["steam", "root", "."]:
            librarycache = os.path.join(steam_dir, folder, "appcache", "librarycache")
            if os.path.exists(librarycache):
                icon_list = glob.glob(librarycache + "/*_icon.jpg", recursive=True)

        self.dbg.stdout("Found {0} Steam icons.".format(len(icon_list)), self.dbg.success, 1)
        return icon_list


class CommonFileTab(TabData):
    """
    Shared class that implements an interactive file management interface for
    features that use a FlatFileManagement() system.

    Each feature (e.g. effects, presets) should re-implement the functions
    where appropriate and specific to different tab interfaces.
    """
    def __init__(self, appdata, filemgr_class, content_widget_name, tree_widget_name):
        """
        Initialise the tab.

        Params:
            appdata                 ApplicationData() object
            filemgr_class           Class to initialise (e.g. effects.EffectFileManagement)
            content_widget_name     Name of QWidget containing the tab data
            tree_widget_name        Name of QTreeWidget containing tasks/file list.
        """
        super().__init__(appdata)

        # Class variables
        self.fileman = filemgr_class()
        self.current_file_data = {}
        self.current_file_path = ""

        # Specific to feature
        self.feature = "unknown"
        self.tab_title = ""
        self.tasks = {
            "example": self._clear_tree
        }

        # Internal codes for UI messages
        self.INFO_NO_ITEMS = 1
        self.INFO_BAD_DATA = 2

        self.Contents = self.main_window.findChild(QWidget, content_widget_name)
        self.Sidebar = self.main_window.findChild(QTreeWidget, tree_widget_name)
        self.TasksBranch = self.Sidebar.invisibleRootItem().child(0)
        self.FilesBranch = self.Sidebar.invisibleRootItem().child(1)

        for branch in [self.TasksBranch, self.FilesBranch]:
            branch.action_id = None
            branch.action_data = None

        self.Sidebar.itemClicked.connect(self._sidebar_changed)

    def _add_tree_item(self, branch, label, icon, action_id, action_data):
        """
        Appends a new QTreeWidgetItem() with the specified parameters.

        action_id and action_data are appended into the object so the
        _sidebar_changed() function knows what to do with the data.
        """
        item = QTreeWidgetItem()
        item.setText(0, label)
        item.setIcon(0, QIcon(icon))
        item.action_id = action_id
        item.action_data = action_data
        branch.addChild(item)

    def _clear_tree(self, branch):
        """
        Clears the contents of a tree branch.
        """
        branch.takeChildren()

    def set_tab(self):
        """
        The feature tab is opened.
        """
        self.set_title(self.tab_title)

        # Open all tree branches
        self.Sidebar.expandAll()

        # Populate files in sidebar
        self._clear_tree(self.FilesBranch)
        file_list = self.fileman.get_item_list()
        for item in file_list:
            self._add_tree_item(self.FilesBranch, item["name"], item["icon"], "open", item["path"])
        self.FilesBranch.sortChildren(0, Qt.AscendingOrder)

        # Show the first item
        if len(file_list) == 0:
            self.show_no_file_screen(0)
        else:
            first_item = self.FilesBranch.child(0)
            first_item.setSelected(True)
            self.open_file(first_item.action_data)

    def _sidebar_changed(self, item):
        """
        A task or file on the sidebar is clicked.

        For ID "tasks", the data is a key that triggers a function. No parameters.
        For ID "open", the data will be opened for the user to work with.
        """
        if item.action_id == "tasks":
            self.Sidebar.selectedItems()[0].setSelected(False)
            self.tasks[item.action_data]()
        elif item.action_id == "open":
            self.open_file(item.action_data)

    def show_no_file_screen(self, message_id):
        """
        The file cannot be opened for viewing - inform the user.
        This needs to be implemented by the feature.

        Params:
            message_id  (int)   One of the self.INFO_* variables
        """
        raise NotImplementedError

    def show_error_message(self, path, error_code):
        """
        The file failed to load. Parse the error code and inform the user.

        Params:
            path        (str)   Path to failed file
            error_code  (int)   Integer from fileman.ERROR_*
        """
        reasons = {
            fileman.ERROR_BAD_DATA: self._("The file contains invalid data."),
            fileman.ERROR_MISSING_FILE: self._("The file no longer exists. Please refresh the page."),
            fileman.ERROR_NEWER_FORMAT: self._("This file was saved in a newer version of this program.")
        }

        try:
            reason = reasons[error_code]
        except KeyError:
            reason = self._("Unspecified error: []").replace("[]", str(error_code))

        self.widgets.open_dialog(self.widgets.dialog_error,
                                 self._("File Error"),
                                 self._("This file cannot be opened:") + "\n" + path,
                                 info_text=reason)

    def _show_file_error(self, traceback=None):
        """
        Show a dialog for the rare scenario where the file could not be written to.
        """
        self.widgets.open_dialog(self.widgets.dialog_error,
                                 self._("File Error"),
                                 self._("The operation could not be completed due to an error processing this file."),
                                 info_text=self._("Please make sure the file permissions are recursively correct:") + '\n' + self.paths.config,
                                 details=traceback)

    def new_file(self):
        """
        Create a new file for this feature.
        The inheriting class should implement this accordingly.
        """
        raise NotImplementedError

    def open_file(self, file_path):
        """
        Opens the specified file in the interface. This should provide an overview
        of the file, a preview and provide options to work with the data.

        The inheriting class should implement this accordingly.
        """
        raise NotImplementedError

    def edit_file(self):
        """
        Open the editor to modify the currently selected file.
        The inheriting class should implement this accordingly.
        """
        raise NotImplementedError

    def delete_file(self):
        """
        Confirm the user is sure the selected file should be deleted.
        """
        name = self.current_file_data["name"]

        def _file_delete_confirmed():
            success = self.fileman.delete_item(self.current_file_path)

            if success != True:
                self._show_file_error(success)
                return

            old_item = self.Sidebar.selectedItems()[0]
            item_index = self.FilesBranch.indexOfChild(old_item)
            self.FilesBranch.removeChild(old_item)
            if self.FilesBranch.childCount() == 0:
                self.show_no_file_screen(0)
                return

            if item_index > self.FilesBranch.childCount() - 1:
                item_index = self.FilesBranch.childCount() - 1

            new_item = self.FilesBranch.child(item_index)
            new_item.setSelected(True)
            self.open_file(new_item.action_data)

        # FIXME: Incomplete features not ready
        msgs = {
            "effects": self._("Permanently delete this effect?\n'[]'\n\nPresets and triggers that use this effect will be unlinked.").replace("[]", name),
            "presets": self._("Permanently delete this preset?\n'[]'\n\nTriggers that use this preset will no longer function.").replace("[]", name),
        }

        msgs = {
            "effects": self._("Permanently delete this effect? \n'[]'").replace("[]", name),
            "presets": self._("Permanently delete this preset? \n'[]'").replace("[]", name),
        }

        self.widgets.open_dialog(self.widgets.dialog_warning,
                                 self._("Confirm Deletion"),
                                 msgs[self.feature],
                                 buttons=[QMessageBox.Yes, QMessageBox.No],
                                 default_button=QMessageBox.Yes,
                                 actions={QMessageBox.Yes: _file_delete_confirmed})

    def clone_file(self):
        """
        Create a new copy of the currently selected effect.
        """
        new_file_path = self.fileman.clone_item(self.current_file_path)

        if not new_file_path:
            self._show_file_error()
            return

        # Get the index for the current item
        current_item = self.Sidebar.selectedItems()[0]
        item_index = self.FilesBranch.indexOfChild(current_item) + 1

        # Add the new item to the list (dynamic, no reloading)
        item = QTreeWidgetItem()
        item.setText(0, self._("[] (Copy)").replace("[]", current_item.text(0)))
        item.setIcon(0, current_item.icon(0))
        item.action_id = "open"
        item.action_data = new_file_path
        self.FilesBranch.insertChild(item_index, item)
        current_item.setSelected(False)
        item.setSelected(True)
        self.open_file(new_file_path)



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
This module is responsible for managing other Polychromatic processes.
"""

import glob
import json
import os
import signal
import shutil
import subprocess

from . import common


class ProcessManager():
    """
    Stores functions for controlling other Polychromatic processes, which may
    be handling a feature (e.g. tray applet) or playing an effect.

    Other Polychromatic processes may send (and be expected to receive) the
    following signals:

    - USR1 "Reload"
      |- Controller           Refresh current device and/or device list.
      |- Tray                 Reload, device list changed.
      |- Helper (FX)          Reload, current effect data changed.
      |- Helper (Triggers)    Reload, trigger data changed.

    - USR2 "Stop"
      |- Controller           (n/a)
      |- Tray                 Exit tray applet
      |- Helper (FX)          Stop playing effect
      |- Helper (Triggers)    Stop monitoring triggers
    """
    def __init__(self, component=None):
        self.pid_dir = self._get_pid_dir()
        self.component = component
        self.components = ["controller", "tray-applet", "helper"]

    def _get_pid_dir(self):
        """
        Returns the runtime directory for storing process PIDs.
        """
        pid_dir = common.paths.pid_dir

        if not os.path.exists(pid_dir):
            os.makedirs(pid_dir)

        return pid_dir

    def _get_pid_file(self):
        """
        Returns the path to a PID file.
        """
        return os.path.join(self.pid_dir, self.component + ".pid")

    def _is_polychromatic_process(self, component, pid):
        """
        Verify that the specified PID is actually an alive Polychromatic process.
        """
        cmdline_file = "/proc/{0}/cmdline".format(pid)

        if os.path.exists(cmdline_file):
            with open(cmdline_file, "r") as f:
                cmdline = f.readline()
        else:
            # No process running here!
            return False

        if cmdline.find("polychromatic-") != -1:
            return True

        # This PID is old and no longer belongs to Polychromatic
        pid_file = os.path.join(self.pid_dir, component + ".pid")
        os.remove(self._get_pid_file())
        return False

    def _get_component_pid(self):
        """
        Returns the PID of a running Polychromatic process, which may be providing
        a feature (e.g. tray applet) or processing software effects for a device.

        Returns:
            (int)       Process ID
            None        Process is not running
        """
        pid_file = self._get_pid_file()

        if not os.path.exists(pid_file):
            return None

        if os.path.getsize(pid_file) == 0:
            return None

        with open(pid_file, "r") as f:
            pid = int(f.read())

        if self._is_polychromatic_process(self.component, pid):
            return pid

        return None

    def _get_component_pid_list(self):
        """
        Returns a list of all the components that currently have a PID file. This
        isn't validated and is presumed to be running and is a Polychromatic process.
        """
        pids = []
        for component in glob.glob(self.pid_dir + "/*.pid"):
            pids.append(os.path.basename(component.replace(".pid", "")))

        return pids

    def set_component_pid(self):
        """
        Assign the PID of the running process to a component or device, indicating
        a 'locked' state to avoid multiple instances.

        If the component is already running, it will be stopped.
        """
        pid_file = self._get_pid_file()

        if os.path.exists(pid_file):
            self.stop()

        with open(pid_file, "w") as f:
            f.write(str(os.getpid()))

        return True

    def release_component_pid(self):
        """
        Unassign the PID of the running process from a component or device.
        """
        pid_file = self._get_pid_file()
        pid = self._get_component_pid()
        if pid == os.getpid():
            os.remove(pid_file)

    def is_another_instance_is_running(self):
        """
        Return a boolean to indicate whether this component is already running.
        """
        pid = self._get_component_pid()
        if pid:
            return True
        return False

    def stop(self):
        """
        Send the USR2 signal to the process to ask this component to stop.
        The PID will be unassigned, allowing it to be used by another
        Polychromatic process.
        """
        pid = self._get_component_pid()
        if pid:
            os.kill(pid, signal.SIGUSR2)

    def reload(self, pid_file=None):
        """
        Send the USR1 signal to reload or restart the component. The PID is
        expected to be reassigned to the new process.
        """
        if not pid_file:
            pid_file = self._get_pid_file()

        pid = self._get_component_pid()
        if pid:
            # Already running, ask to restart self
            os.kill(pid, signal.SIGUSR1)
        else:
            # Not running, start!
            self.start_component()

    def _get_component_exec_path(self, name):
        """
        Internally gets the path of a component, such as "tray-applet" or "controller".

        Returns:
            (str)   Path to exceutable, e.g. "/usr/bin/polychromatic-controller"
            None    Executable not found.
        """
        bin_filename = "polychromatic-" + name

        # __file__ = procpid.py
        path_relative = os.path.abspath(os.path.join(os.path.dirname(__file__), "../", bin_filename))
        path_var = shutil.which(bin_filename)
        path_system = "/usr/bin/" + bin_filename

        for path in [path_relative, path_var, path_system]:
            if path and os.path.exists(path):
                return path

        return None

    def start_component(self, parameters=[]):
        """
        Start a new Polychromatic process with 'name' being the suffix of the
        executable, e.g. 'tray-applet'

        Returns a boolean to indicate whether the process successfully spawned.
        """
        if self.component not in self.components:
            return False

        bin_path = self._get_component_exec_path(self.component)

        try:
            bin_arguments = [bin_path]
            bin_arguments += parameters
            subprocess.Popen(bin_arguments, env=dict(os.environ))
            return True
        except Exception as e:
            print("Failed to start process: " + " ".join(bin_arguments))
            print(str(e))
            return False

        print("Could not locate executable: polychromatic-" + self.component)
        return False

    def is_component_installed(self, name):
        """
        Checks whether another Polychromatic component is installed. This could be
        because of modular packaging, where a user installs the tray applet, but
        not the Controller, for example.

        Returns a boolean to indicate it is launchable.
        """
        if self._get_component_exec_path(name):
            return True
        return False

    def restart_self(self, exec_path, exec_args):
        """
        Immediately restart the current execution.
        """
        os.execv(exec_path, exec_args)

    def restart_all(self):
        """
        Restart all tasks, excluding the current process.
        """
        for pid_file in self._get_component_pid_list():
            procmgr = ProcessManager(pid_file)
            procmgr.reload(pid_file)


class DeviceSoftwareState(object):
    """
    Tracks the active custom software effect or preset for a specified device,
    such as when a custom effect is currently being played and/or a preset is
    currently in use.

    This is metadata presented to the user and internally to know the last
    effect or preset (when autostarting)

    The "state" is stored in a JSON file named after the device's serial number.
    {
        "effect": {
            "name": "Human readable name",
            "icon": "<absolute path>",
            "path": "<path to effect JSON>"
        },
        "preset": {
            "name": "Human readable name",
            "icon": "<absolute path>",
            "path": "<path to preset JSON>"
        }
    }
    """
    def __init__(self, serial):
        self.serial = serial
        self.state = {}
        self.state_path = os.path.join(common.paths.states, serial + ".json")

        if not os.path.exists(self.state_path):
            with open(self.state_path, "w") as f:
                f.write("{}")

        self._read_state()

    def _read_state(self):
        with open(self.state_path) as f:
            try:
                self.state = json.load(f)
            except Exception:
                # Bad JSON or filesystem error. Ignore and start afresh.
                print("Ignoring bad data: ", self.state_path)
                self.state = {}

    def _write_state(self):
        with open(self.state_path, "w") as f:
            f.write(json.dumps(self.state))

    def get_preset(self):
        """
        Returns the metadata of the preset that is currently representing this
        device's current status.

        If no preset is set, this will return None.
        """
        try:
            return {
                "name": self.state["preset"]["name"],
                "icon": self.state["preset"]["icon"],
                "path": self.state["preset"]["path"]
            }
        except KeyError:
            return None

    def set_preset(self, name, icon, path):
        """
        This device is now aligned to the properties of a saved preset.
        """
        self.state["preset"] = {}
        self.state["preset"]["name"] = name
        self.state["preset"]["icon"] = icon
        self.state["preset"]["path"] = path
        self._write_state()

    def clear_preset(self):
        """
        This device no longer matches the preset state.
        """
        try:
            del(self.state["preset"])
            self._write_state()
        except KeyError:
            # Does not exist.
            pass

    def get_effect(self):
        """
        Returns the metadata of the effect that is currently playing on this
        device. This only applies to custom software effects and not any
        backend or hardware effects.

        If no effect is running, this will return None.
        """
        try:
            return {
                "name": self.state["effect"]["name"],
                "icon": self.state["effect"]["icon"],
                "path": self.state["effect"]["path"]
            }
        except KeyError:
            return None

    def set_effect(self, name, icon, path):
        """
        This device is now running under a software effect.
        """
        self.state["effect"] = {}
        self.state["effect"]["name"] = name
        self.state["effect"]["icon"] = icon
        self.state["effect"]["path"] = path
        self._write_state()

    def clear_effect(self):
        """
        This device is no longer under software control.
        """
        try:
            del(self.state["effect"])
            self._write_state()
        except KeyError:
            # Does not exist.
            pass



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2016-2024 Luke Horwell <code@horwell.me>
"""
Module for determining where files are located.
"""
import os
import sys


class Paths():
    """
    Reference file/folder paths for data files, configuration and cache directories.
    """
    def __init__(self):
        # System-wide data
        self.data_dir = self.get_data_path()

        # Local data
        self.config = self.get_config_path()
        self.cache = self.get_cache_path()
        self.dev = self.set_dev_mode()
        self.pid_dir = self.get_pid_path()

        # Caches
        self.assets_cache = os.path.join(self.cache, "assets")
        self.effects_cache = os.path.join(self.cache, "effects")
        self.webview_cache = os.path.join(self.cache, "editor")

        # Save Data
        self.devices = os.path.join(self.config, "devices")
        self.dpi = os.path.join(self.config, "dpi")
        self.effects = os.path.join(self.config, "effects")
        self.presets = os.path.join(self.config, "presets")
        self.custom_icons = os.path.join(self.config, "custom_icons")
        self.states = os.path.join(self.config, "states")

        # Files
        self.preferences = os.path.join(self.config, "preferences.json")
        self.colours = os.path.join(self.config, "colours.json")

        # Legacy (<= v0.3.12)
        self.old_profile_folder = os.path.join(self.config, "profiles")
        self.old_profile_backups = os.path.join(self.config, "backups")
        self.old_devicestate = os.path.join(self.config, "devicestate.json")

        self.create_dirs_if_not_exist()

    def create_dirs_if_not_exist(self):
        """
        Ensure all the directories exist for the application.
        """
        for folder in [self.config, self.presets, self.custom_icons, self.states, self.devices, self.effects,
                       self.cache, self.assets_cache, self.effects_cache, self.webview_cache, self.dpi]:
            if not os.path.exists(folder):
                os.makedirs(folder)

    def set_dev_mode(self):
        """
        When developing within the repository, change the paths accordingly.
        """
        try:
            if os.environ["POLYCHROMATIC_DEV_CFG"] == "true":
                # __file__ = polychromatic/base.py
                self.cache = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "savedatadev", "cache"))
                self.config = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "savedatadev", "config"))
        except KeyError:
            return False
        return True

    @staticmethod
    def get_data_path():
        """
        For development/opt, this is normally adjacent to the application executable.
        For system-wide installs, this is generally /usr/share/polychromatic.
        """
        module_path = __file__

        if os.path.exists(os.path.join(os.path.dirname(module_path), "../data/img/")):
            return os.path.abspath(os.path.join(os.path.dirname(module_path), "../data/"))

        for directory in ["/usr/local/share/polychromatic", "/usr/share/polychromatic"]:
            if os.path.exists(directory):
                return directory

        print("Cannot locate data directory! Please reinstall the application.")
        sys.exit(1)

    @staticmethod
    def get_config_path():
        """
        Path for persistent save data for the application.
        """
        try:
            return os.path.join(os.environ["XDG_CONFIG_HOME"], "polychromatic")
        except KeyError:
            return os.path.join(os.path.expanduser("~"), ".config", "polychromatic")

    @staticmethod
    def get_cache_path():
        """
        Path for temporary data to speed up processing later.
        """
        try:
            return os.path.join(os.environ["XDG_CACHE_HOME"], "polychromatic")
        except KeyError:
            return os.path.join(os.path.expanduser("~"), ".cache", "polychromatic")

    @staticmethod
    def get_pid_path():
        """
        Runtime directory for PID text files that reference other Polychromatic processes.
        """
        try:
            return os.path.join(os.environ["XDG_RUNTIME_DIR"], "polychromatic")
        except KeyError:
            return "/tmp/polychromatic"



from polychromatic.backends._backend import Backend as Backend

import os


class DummyMatrix(Backend.DeviceItem.Matrix):
    """
    A simulated matrix that does absolutely nothing.
    """
    def __init__(self, *args):
        pass

    def set(self, x, y, r, g, b):
        pass

    def draw(self):
        pass

    def clear(self):
        pass


class DummyDPI(Backend.DeviceItem.DPI):
    """
    Simulate DPI tracked using a text file.
    """
    def __init__(self, serial):
        super().__init__()
        self.x = 800
        self.y = 1800
        self.min = 100
        self.max = 16000

        self.path = os.path.expanduser(f"~/.cache/polychromatic/test_dpi_{serial}")

        # Write initial DPI files
        dpi_dir = os.path.dirname(self.path)
        if not os.path.exists(dpi_dir):
            os.makedirs(dpi_dir)
        self.set(self.x, self.y)

    def refresh(self):
        with open(self.path, "r") as f:
            self.x = int(f.readline())
            self.y = int(f.readline())

    def set(self, x, y):
        with open(self.path, "w") as f:
            f.write(str(self.x) + "\n")
            f.write(str(self.y) + "\n")


class DummyBackend(Backend):
    """
    Simulates an imaginary backend with a few devices for testing logic.

    Refer to the docstrings for the specifications for each device. There are
    some 'unknown devices' too, for a scenario where the device is not supported,
    or not set up properly.
    """
    def __init__(self, *args):
        super().__init__(*args)
        self.backend_id = "dummy"
        self.logo = "polychromatic.svg"
        self.version = "9.9.9"
        self.project_url = "https://polychromatic.app"
        self.bug_url = "https://github.com/polychromatic/polychromatic/issues"
        self.releases_url = "https://github.com/polychromatic/polychromatic/releases"
        self.license = "GPLv3"

    def init(self):
        return True

    def get_unsupported_devices(self):
        unknown_devices = []
        for i in range(0, 3):
            device = Backend.UnknownDeviceItem()
            device.name = "Unknown Device " + str(i)
            device.form_factor = self.get_form_factor()
            unknown_devices.append(device)
        return unknown_devices

    def _get_keyboard_device(self):
        """
        - 1 Zone
        - RGB Support
        - Implements every option type/combo:
            - None          EffectOption
            - Static        EffectOption with colour only
            - Wave          EffectOption with parameters only               <<<
        """
        device = Backend.DeviceItem()
        device.name = "Dummy Keyboard"
        device.form_factor = self.get_form_factor("keyboard")
        device.serial = "DUMMY0001"
        device.keyboard_layout = "en_GB"
        device.matrix = DummyMatrix()

        zone = Backend.DeviceItem.Zone()
        zone.zone_id = "main"
        zone.label = "Main Zone"
        device.zones.append(zone)

        class Nothing(Backend.EffectOption):
            def __init__(self):
                super().__init__()
                self.uid = "none"
                self.label = "None"

            def apply(self):
                pass

        class Static(Backend.EffectOption):
            def __init__(self):
                super().__init__()
                self.uid = "static"
                self.label = "Static"
                self.active = True
                self.colours_required = 1
                self.colours = ["#00FF00"]

            def apply(self):
                pass

        class Wave(Backend.EffectOption):
            def __init__(self):
                super().__init__()
                self.uid = "wave"
                self.label = "Wave"

                param_1 = Backend.EffectOption.Parameter()
                param_1.data = 1
                param_1.label = "Left"

                param_2 = Backend.EffectOption.Parameter()
                param_2.data = 2
                param_2.label = "Right"
                param_2.active = True

                self.parameters = [param_1, param_2]

            def apply(self, data):
                pass

        for option in [Nothing, Static, Wave]:
            zone.options.append(option())

        return device

    def _get_mouse_device(self):
        device = Backend.DeviceItem()
        device.name = "Dummy Mouse"
        device.form_factor = self.get_form_factor("mouse")
        device.serial = "DUMMY0002"
        device.dpi = DummyDPI(device.serial)
        return device

    def _get_headset_device(self):
        device = Backend.DeviceItem()
        device.name = "Dummy Headset"
        device.form_factor = self.get_form_factor("headset")
        device.serial = "DUMMY0003"
        return device

    def get_devices(self):
        return [
            self._get_keyboard_device(),
            self._get_mouse_device(),
            self._get_headset_device(),
        ]

    def get_device_by_name(self, name):
        names = {
            "Dummy Keyboard": self._get_keyboard_device,
            "Dummy Mouse": self._get_mouse_device,
            "Dummy Headset": self._get_headset_device,
        }
        return names[name]()

    def get_device_by_serial(self, serial):
        serials = {
            "DUMMY0001": self._get_keyboard_device,
            "DUMMY0002": self._get_mouse_device,
            "DUMMY0003": self._get_headset_device,
        }
        return serials[serial]()

    def troubleshoot(self):
        return None

    def restart(self):
        return True



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2022 Luke Horwell <code@horwell.me>
"""
Handles the bulk "Apply to All" options to apply settings to all devices at once.
"""
from .backends._backend import Backend
from . import common
from . import middleman as mn
from . import preferences


class BulkOption(object):
    """
    An object representing a selectable 'apply to all' button.
    """
    def __init__(self, options=[], middleman=mn.Middleman, label="", icon="", value=None):
        self.options = options
        self.middleman = middleman
        self.label = label
        self.icon = icon
        self.value = value

    def apply(self):
        raise NotImplementedError


class _BulkBrightness(BulkOption):
    def apply(self, a=None):
        # 'a' is an object passed from Tray Applet. Unnecessary.
        for option in self.options:
            if isinstance(option, Backend.SliderOption):
                option.apply(self.value)
            elif isinstance(option, Backend.ToggleOption):
                option.apply(True if self.value > 0 else False)


class _BulkEffect(BulkOption):
    def apply(self, a=None):
        # 'a' is an object passed from Tray Applet. Unnecessary.
        for option in self.options:
            if not option.uid == self.value:
                continue
            if option.parameters:
                default_param = self.middleman.get_default_parameter(option)
                option.apply(default_param.data)
            else:
                option.apply()


class _BulkColour(BulkOption):
    def apply(self, a=None):
        # 'a' is an object passed from Tray Applet. Unnecessary.
        for device in self.options:
            device.refresh()
            try:
                self.middleman.set_colour_for_active_effect_device(device, self.value)
            except IndexError:
                # Not supported for this effect
                pass


class BulkApplyOptions(object):
    """
    Assembles and processes "Apply to All" options for applying a brightness,
    effect or colour to all devices at once. The Controller and tray applet
    interfaces will interpret this data.
    """
    def __init__(self, middleman=mn.Middleman):
        self.middleman = middleman
        self.devices = middleman.get_devices()

        # Stores BulkOption() objects
        self.brightness = []
        self.effects = []
        self.colours = []

        # Some effects not available for every device?
        self.mix_match = False

        self.refresh()

    def refresh(self):
        """
        Analyze devices and populate the bulk objects.
        """
        brightness = []
        effects = []
        has_colours = False

        for device in self.devices:
            for zone in device.zones:
                for option in zone.options:
                    if option.uid == "brightness":
                        brightness.append(option)
                    elif isinstance(option, Backend.EffectOption):
                        effects.append(option)
                    if option.colours_required > 0:
                        has_colours = True

        self._populate_bulk_brightness(brightness)
        self._populate_bulk_effects(effects)
        self._populate_bulk_colours(has_colours)

    def _populate_bulk_brightness(self, options):
        """
        Builds new BulkOption() objects that apply a brightness across all devices and zones.

        NOTE: The values in this function are hardcoded for OpenRazer (0-100). More work
        is required if another backend handles 'brightness' differently (e.g. 0-255)
        """
        self.brightness = []

        for value in [0, 25, 50, 75, 100]:
            label = str(value) + "%"
            icon = common.get_icon("params", str(value))
            self.brightness.append(_BulkBrightness(options, self.middleman, label, icon, value))

    def _populate_bulk_effects(self, options):
        """
        Builds new BulkOption() objects that apply an effect across all devices and zones.
        """
        self.effects = []
        uids = []
        occurrences = {}
        effects = {}

        for option in options:
            try:
                occurrences[option.uid] += 1
            except KeyError:
                occurrences[option.uid] = 1

            # Already added item?
            if option.uid in uids:
                continue

            label = option.label
            icon = option.icon
            effects[option.uid] = _BulkEffect(options, self.middleman, label, icon, option.uid)
            uids.append(option.uid)

        # If the effect can't apply to all devices, add an asterisk
        total_devices = len(self.devices)
        for uid in occurrences.keys():
            if occurrences[uid] < total_devices:
                effects.get(uid).label += "*"
                self.mix_match = True

        for option in effects.keys():
            self.effects.append(effects[option])

    def _populate_bulk_colours(self, has_colours):
        """
        Builds new BulkOption() objects that apply a colour across all devices and zones.
        """
        self.colours = []

        if not has_colours:
            return

        # FIXME: Refactoring required: Not passing _
        colours = preferences.get_colour_list(None)
        for colour in colours:
            colour_hex = colour["hex"]
            label = colour["name"]

            # FIXME: Finish refactoring for function
            icon = common.generate_colour_bitmap(None, colour_hex)
            self.colours.append(_BulkColour(self.devices, self.middleman, label, icon, colour_hex))




import polychromatic.common as common
import polychromatic.locales as locales
import polychromatic.preferences as preferences
import polychromatic.effects as effects
import polychromatic.fileman as fileman

import glob
import json
import os
import unittest


class TestEffects(unittest.TestCase):
    """
    Test the internals of Polychromatic.
    """
    @classmethod
    def setUpClass(self):
        self.i18n = locales.Locales("polychromatic")
        self._ = self.i18n.init()
        self.dbg = common.Debugging()
        self.paths = common.paths
        preferences.init(self._)

        self.fileman = effects.EffectFileManagement()

        # Dummy content
        self.fileman.save_item(self.fileman.init_data("Test 1", effects.TYPE_LAYERED))
        self.fileman.save_item(self.fileman.init_data("Test 2", effects.TYPE_LAYERED))
        self.fileman.save_item(self.fileman.init_data("Test 3", effects.TYPE_LAYERED))

        # Unit test resources
        self.res_path = os.path.join(os.path.dirname(__file__), "files")

    @classmethod
    def tearDownClass(self):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_create_layered_effect(self):
        data1 = self.fileman.init_data("My Layered Effect", effects.TYPE_LAYERED)
        success, path = self.fileman.save_item(data1)
        data2 = self.fileman.get_item(path)
        self.assertEqual(type(data2), dict, "Could not create a layered effect")

    def test_create_scripted_effect(self):
        data1 = self.fileman.init_data("My Scripted Effect", effects.TYPE_SCRIPTED)
        success, path = self.fileman.save_item(data1)
        data2 = self.fileman.get_item(path)
        self.assertEqual(type(data2), dict, "Could not create a scripted effect")

    def test_create_sequence_effect(self):
        data1 = self.fileman.init_data("My Sequence Effect", effects.TYPE_SEQUENCE)
        success, path = self.fileman.save_item(data1)
        data2 = self.fileman.get_item(path)
        self.assertEqual(type(data2), dict, "Could not create a sequence effect")

    def test_illegal_unix_filename(self):
        data = self.fileman.init_data("Game / Application & More", effects.TYPE_LAYERED)
        success, path = self.fileman.save_item(data)
        self.assertTrue(os.path.exists(path), "Failed to strip illegal Unix filename characters")

    def test_unicode_characters(self):
        data = self.fileman.init_data("Game: What? Who! Where: ƀƁƂƃƄƅƆƇƈƉƊƌƍƎƏƐƑƒ 😁", effects.TYPE_LAYERED)
        success, path = self.fileman.save_item(data)
        self.assertTrue(os.path.exists(path), "Failed to process Unicode characters")

    def test_get_effect_list(self):
        items = self.fileman.get_item_list()
        self.assertGreaterEqual(len(items), 3, "Could not get a list of effects")

    def test_get_effect(self):
        items = self.fileman.get_item_list()
        data = self.fileman.get_item(items[0]["path"])
        self.assertEqual(type(data["name"]), str, "Could not get an effect")

    def test_delete_effect(self):
        items = self.fileman.get_item_list()
        data = self.fileman.get_item(items[0]["path"])
        path_to_delete = data["parsed"]["path"]
        self.fileman.delete_item(path_to_delete)
        self.assertEqual(os.path.exists(path_to_delete), False, "Could not delete an effect")

    def test_clone_effect(self):
        items = self.fileman.get_item_list()
        data = self.fileman.get_item(items[0]["path"])
        source_path = data["parsed"]["path"]
        cloned_path = self.fileman.clone_item(source_path)
        self.assertEqual(os.path.exists(source_path), os.path.exists(cloned_path), "Could not clone an effect")

    def test_rename_effect(self):
        items = self.fileman.get_item_list()
        data = self.fileman.get_item(items[0]["path"])
        orig_path = data["parsed"]["path"]
        data["name"] = "NEW NAME"
        self.fileman.save_item(data, orig_path)
        self.assertEqual(os.path.exists(orig_path), False, "Could not rename an effect")

    def test_saving_duplicate_name(self):
        items = self.fileman.get_item_list()
        data = self.fileman.get_item(items[0]["path"])
        orig_path = data["parsed"]["path"]
        success, new_path = self.fileman.save_item(data)
        self.assertNotEqual(orig_path, new_path, "Could not save an effect with a duplicate filename")

    def test_devicemap_map_json(self):
        passed = True
        with open("data/devicemaps/maps.json") as f:
            data = json.load(f)
        for name in data.keys():
            item = data[name]
            keys = list(item.keys())
            if False in [
                "filename" in keys,
                "rows" in keys,
                "cols" in keys,
                "locale" in keys,
                "scancode" in keys
            ]:
                passed = False

        self.assertEqual(passed, True, "Item(s) missing keys in maps.json")

    def test_devicemap_valid_json(self):
        passed = True
        for svg_file in glob.glob("data/devicemaps/*.json"):
            with open(svg_file) as f:
                try:
                    data = json.load(f)
                except Exception as e:
                    print(svg_file, str(e))
                    passed = False

        self.assertEqual(passed, True, "Invalid JSON: " + svg_file)

    def test_devicemap_map_exists(self):
        passed = True
        with open("data/devicemaps/maps.json") as f:
            data = json.load(f)
        for name in data.keys():
            if not os.path.exists("data/devicemaps/" + data[name]["filename"]):
                passed = False

        self.assertEqual(passed, True, "maps.json referenced SVG files that do not exist")

    def test_devicemap_map_xy_check(self):
        for svg_file in glob.glob("data/devicemaps/*.svg"):
            with open(svg_file) as f:
                data = " ".join(f.readlines())

                # SVGs should have at least one "LED" class.
                if data.find("LED") == -1:
                    self.assertTrue(False, "Device map {0} does not specify an LED class!".format(os.path.basename(svg_file)))
                found_an_ID = False
                for x in range(0, 50):
                    for y in range(0, 50):
                        if data.find("x{0}-y{1}".format(str(x), str(y))) != -1:
                            found_an_ID = True

                # SVGs should have at least one "x0-y0" ID.
                if not found_an_ID:
                    self.assertTrue(False, "Device map {0} does not specify IDs!".format(os.path.basename(svg_file)))

        self.assertEqual(True, True)



#!/usr/bin/python3
#
# This script is used by the CI system to produce changelogs for automated
# Debian/Ubuntu builds.
#
# Parameters: <release> <codename>
# E.g.        "20.04" "focal"
#

import datetime
import os
import subprocess
import sys

UBUNTU_RELEASE = sys.argv[1]
UBUNTU_CODENAME = sys.argv[2]
CHANGELOG = os.path.join(os.path.dirname(__file__), "../../debian/changelog")
PKG_VERSION = "0.0.0"


def run_command(cmd):
    cmd = cmd.split(" ")
    return str(subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0].decode("UTF-8")).strip()


# Determine package version
ver_raw = run_command("git describe --tags")[1:]
ver_parts = ver_raw.split("-")
if len(ver_parts) == 3:
    # Example: 0.6.0-41-g59319b6
    PKG_VERSION = "{0}-{1}~{2}~{3}".format(ver_parts[0], ver_parts[1], ver_parts[2], UBUNTU_RELEASE)
    START_TAG = "v" + ver_parts[0]
else:
    # Example: 1.0.0
    PKG_VERSION = ver_raw
    START_TAG = "v" + ver_raw


# Generate a changelog to describe latest changes
LOG = run_command("git log {0}..HEAD --oneline".format(START_TAG)).strip()
SIGN_DATE = datetime.datetime.now().strftime("%a, %d %b %Y %H:%M:%S")

def get_changelog():
    lines = []
    lines.append("polychromatic ({0}) {1}; urgency=low\n\n".format(PKG_VERSION, UBUNTU_CODENAME))
    lines.append("  This package was automatically generated.\n\n")
    if len(LOG.split("\n")) > 1:
        lines.append("  Commit changes (newest first):\n\n")
        for line in LOG.split("\n"):
            lines.append("  * " + line + "\n")
    lines.append("\n -- Polychromatic Builder <bot@polychromatic.app>  {0} +0000\n\n".format(SIGN_DATE))

    return "".join(lines)


# Generate and upload each package
with open(CHANGELOG, "w") as f:
    f.writelines(get_changelog())



from _dummy import DummyBackend as DummyBackend

import polychromatic.base as base
import polychromatic.common as common
import polychromatic.locales as locales
import polychromatic.middleman as middleman
import polychromatic.preferences as preferences

import os
import unittest


class TestMiddleman(unittest.TestCase):
    """
    Test the middleman against a dummy module to check data flows as expected.
    """
    @classmethod
    def setUpClass(self):
        # FIXME: Middleman is a mess, skip tests
        self.skipTest(self, "Skipped")

        # Override all modules with the dummy
        self.base = base.PolychromaticBase()
        self.base.init_base("", [])
        self.middleman = middleman.Middleman()
        self.middleman._base = self.base
        self.middleman.init()

        if not self.middleman.backends:
            raise RuntimeError("Could not init dummy module!")

    @classmethod
    def tearDownClass(self):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_backend_running(self):
        self.assertTrue(self.middleman.is_backend_running("dummy"))

    def test_versions(self):
        self.assertTrue(self.middleman.get_versions()["dummy"], "9.9.9")

    def test_get_devices(self):
        self.assertGreater(len(self.middleman.get_devices()), 0)

    def test_device_cache(self):
        item = self.middleman.get_devices()[0].name = "POSION"
        self.middleman.reload_device_cache()
        self.assertNotEqual(self.middleman.get_devices()[0].name, "POSION")

    def test_get_device_by_name(self):
        device = self.middleman.get_device_by_name("Dummy Headset")
        self.assertEqual(device.serial, "DUMMY0003")

    def test_get_device_by_serial(self):
        device = self.middleman.get_device_by_serial("DUMMY0002")
        self.assertEqual(device.name, "Dummy Mouse")

    def test_get_devices_by_form_factor(self):
        device = self.middleman.get_devices_by_form_factor("keyboard")[0]
        self.assertEqual(device.name, "Dummy Keyboard")

    def test_unsupported_devices(self):
        unknown = self.middleman.get_unsupported_devices()
        self.assertEqual(len(unknown), 3)

    def test_troubleshoot(self):
        # TODO: Incomplete!
        self.skipTest("")

    def test_restart(self):
        self.assertTrue(self.middleman.restart("dummy"))

    def test_active_effect(self):
        # The dummy keyboard's active effect is Static
        device = self.middleman.get_device_by_name("Dummy Keyboard")
        zone = device.zones[0]
        self.assertEqual(self.middleman.get_active_effect(zone), zone.options[1])

    def test_replay_active_effect(self):
        device = self.middleman.get_device_by_name("Dummy Keyboard")
        self.middleman.replay_active_effect(device)

    def test_set_colour_for_option(self):
        device = self.middleman.get_device_by_name("Dummy Keyboard")
        option = device.zones[0].options[1]
        self.middleman.set_colour_for_option(option, "#FF0000")
        self.assertEqual(option.colours[0], "#FF0000")

    def set_colour_for_active_effect_zone(self):
        device = self.middleman.get_device_by_name("Dummy Keyboard")
        expected_option = device.zones[0].options[1]
        zone = device.zones[0]
        self.middleman.set_colour_for_active_effect_zone(zone, "#0000FF")
        self.assertEqual(expected_option.colours[0], "#0000FF")



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2015-2024 Luke Horwell <code@horwell.me>
"""
This module is responsible for loading/saving persistent data used by Polychromatic's frontends.
"""

import os
import glob
import json
import shutil
from . import common
from . import locales

VERSION = 8

TOOLBAR_STYLE_DEFAULT = 0
TOOLBAR_STYLE_ICONS_ONLY = 1
TOOLBAR_STYLE_TEXT_ONLY = 2
TOOLBAR_STYLE_ALONGSIDE = 3
TOOLBAR_STYLE_UNDER = 4

LANDING_TAB_DEVICES = 0
LANDING_TAB_EFFECTS = 1
# LANDING_TAB_PRESETS = 2
# LANDING_TAB_TRIGGERS = 3

WINDOW_BEHAVIOUR_CENTER = 0
WINDOW_BEHAVIOUR_REMEMBER = 1
WINDOW_BEHAVIOUR_IGNORE = 2
WINDOW_BEHAVIOUR_MAXIMIZED = 3

TRAY_AUTO = 0
TRAY_APPINDICATOR = 1
TRAY_GTK_STATUS = 2
TRAY_AYATANA = 3

dbg = common.Debugging()
path = common.paths


def load_file(filepath):
    """
    Loads a JSON file from disk. If empty, it will be created.

    Params:
        filepath            String from the Path() object.

    Returns:
        {}                  Data (dictionary object)
    """
    if not os.path.exists(filepath):
        init_config(filepath)
        data = {}

    try:
        with open(filepath) as stream:
            data = json.load(stream)
    except Exception as e:
        dbg.stdout(filepath + ": Read error!", dbg.error)
        dbg.stdout("Exception:\n" + common.get_exception_as_string(e), dbg.error)
        init_config(filepath)
        data = {}

    # Check preferences contain valid data and defaults.
    def _validate(group, item, data_type, default_value):
        try:
            data[group]
        except KeyError:
            data[group] = {}

        try:
            data[group][item]
        except KeyError:
            data[group][item] = default_value
            save_file(path.preferences, data)

        if type(data[group][item]) != data_type:
            data[group][item] = default_value
            save_file(path.preferences, data)

    if filepath == path.preferences:
        # (group, item, type, default)
        _validate("editor", "live_preview", bool, True)
        _validate("editor", "hide_key_labels", bool, False)
        _validate("editor", "system_cursors", bool, False)
        _validate("editor", "suppress_confirm_dialog", bool, False)
        _validate("editor", "show_saved_colour_shades", bool, True)
        _validate("controller", "download_device_images", bool, True)
        _validate("controller", "landing_tab", int, 0)
        _validate("controller", "show_menu_bar", bool, True)
        _validate("controller", "system_qt_theme", bool, False)
        _validate("controller", "window_behaviour", int, 0)
        _validate("controller", "toolbar_style", int, TOOLBAR_STYLE_ALONGSIDE)
        _validate("tray", "autostart", bool, True)
        _validate("tray", "mode", int, 0)
        _validate("tray", "icon", str, common.get_default_tray_icon())
        _validate("tray", "autostart_delay", int, 0)

        for prefix in ["main", "editor"]:
            _validate("geometry", prefix + "_window_pos_x", int, 0)
            _validate("geometry", prefix + "_window_pos_y", int, 0)
            _validate("geometry", prefix + "_window_size_x", int, 1000)
            _validate("geometry", prefix + "_window_size_y", int, 600)

    return(data)


def save_file(filepath, newdata):
    """
    Commit data to the disk.

    Params:
        filepath            String from the Path() object.
        newdata             Data (dictionary object)

    Returns:
        True                Save successful.
        False               Save failed.
    """
    # The preferences file always stores the configuration version.
    if filepath == path.preferences:
        newdata["config_version"] = VERSION

    if not os.path.exists(filepath):
        open(filepath, "w").close()

    if os.access(filepath, os.W_OK):
        with open(filepath, "w+") as f:
            f.write(json.dumps(newdata, sort_keys=True, indent=4))
        return True
    else:
        return False


def init_config(filepath):
    """
    Prepares a configuration file for the first time.
    """
    try:
        # Backup if the JSON was invalid.
        if os.path.exists(filepath):
            dbg.stdout(filepath + ": JSON corrupt or unreadable. It will backed up then recreated.", dbg.error)
            os.rename(filepath, filepath + ".bak")

        # Touch new file
        save_file(filepath, {})
        dbg.stdout(filepath + ": New configuration written.", dbg.success, 1)

    except Exception as e:
        # Couldn't create the default configuration.
        dbg.stdout(filepath + ": Init write error!", dbg.error)
        dbg.stdout("Exception: " + str(e), dbg.error)


def upgrade_old_pref():
    """
    Checks and updates the configuration from previous revisions.
    """
    try:
        with open(path.preferences, "r") as f:
            data = json.load(f)
        config_version = int(data["config_version"])
    except Exception:
        # Never mind, the parent function should fix this later.
        return

    # Is the configuration version already up-to-date?
    if VERSION == config_version:
        return

    # Is the config newer then the software? Wicked time travelling!
    if config_version > VERSION:
        dbg.stdout("\nWARNING: Your preferences file is newer then the application!", dbg.error)
        dbg.stdout("It's likely you're running an older version. This is unsupported.", dbg.error)
        dbg.stdout("     Current Config Version:   v." + str(config_version), dbg.error)
        dbg.stdout("     Installed Config Version: v." + str(VERSION), dbg.error)
        dbg.stdout("")
        return

    # Always clear cache when configuration is updated
    if os.path.exists(path.cache):
        shutil.rmtree(path.cache)
        os.makedirs(path.cache)
        dbg.stdout("Cache cleared.", dbg.success, 1)

    dbg.stdout("Upgrading configuration from v{0} to v{1}...".format(config_version, VERSION), dbg.action)

    # v0.3.12
    if config_version < 5:
        # Ensure preferences.json is clean.
        data = load_file(path.preferences)
        for key in ["activate_on_save", "live_switch", "live_preview"]:
            try:
                value = data["editor"][key]
                if type(value) == str:
                    if value in ['true', 'True']:
                        data["editor"][key] = True
                    else:
                        data["editor"][key] = False
            except Exception:
                pass

        save_file(path.preferences, data)

    # v0.4.0 (dev)
    if config_version == 6:
        # The configuration will be reset.
        dbg.stdout("Unsupported v0.4.0 configuration detected. Preferences reset.", dbg.warning)
        for filepath in [path.preferences, path.colours, path.old_devicestate]:
            if os.path.exists(filepath):
                os.remove(filepath)
        return

    # v0.5.0 (dev)
    if config_version < 7:
        # Migrate preferences.json to new keys
        old_data = load_file(path.preferences)

        # -- Tray icon is now one key (a relative or absolute path)
        new_tray_value = ""

        def get_path_from_gtk_icon_name(icon_name):
            """
            Returns an image path determined by a GTK icon name, if there is one.

            This is a legacy feature as the tray icon is a relative/absolute path,
            but remains so users upgrading from v0.3.12.
            """
            import gi
            gi.require_version("Gtk", "3.0")
            from gi.repository import Gtk

            theme = Gtk.IconTheme.get_default()
            info = theme.lookup_icon(icon_name, 22, 0)
            try:
                filename = info.get_filename()
            except Exception:
                filename = None

            if filename:
                return filename
            else:
                return ""

        try:
            old_type = old_data["tray_icon"]["type"]
            if old_type == "gtk":
                new_tray_value = get_path_from_gtk_icon_name(old_data["tray_icon"]["value"])
            elif old_type == "custom":
                new_tray_value = old_data["tray_icon"]["value"]
            elif old_type == "builtin":
                try:
                    mapping = {
                        "0": "img/tray/light/humanity.svg",
                        "1": "img/tray/dark/humanity.svg",
                        "2": "img/tray/animated/chroma.gif",
                        "3": "img/tray/light/breeze.svg",
                        "4": "img/tray/dark/breeze.svg"
                    }
                    new_tray_value = mapping[old_data["tray_icon"]["value"]]
                except KeyError:
                    # Invalid data, discard.
                    pass
        except ImportError:
            dbg.stdout("GTK not installed. Tray icon cannot be migrated.", dbg.error)
        except KeyError:
            # Invalid data, discard.
            pass

        try:
            old_live_preview = old_data["editor"]["live_preview"]
        except KeyError:
            old_live_preview = False

        new_data = {
            "colours": {
                "primary": "#00FF00",               # New
                "secondary": "#00FFFF"              # New
            },
            "editor": {
                "live_preview": old_live_preview    # Changed
            },
            "tray": {
                "icon": new_tray_value              # Changed
            }
        }

        os.remove(path.preferences)
        save_file(path.preferences, new_data)

        # devicestate.json now obsolete
        if os.path.exists(path.old_devicestate):
            os.remove(path.old_devicestate)

        # If the colours were unchanged from v0.3.12, reset to new ones.
        old_colours = load_file(path.colours)
        old_colour_json = {
            "1": {"name": "White", "col": [255, 255, 255]},
            "2": {"name": "Red", "col": [255, 0, 0]},
            "3": {"name": "Orange", "col": [255, 165, 0]},
            "4": {"name": "Yellow", "col": [255, 255, 0]},
            "5": {"name": "Signature Green", "col": [0, 255, 0]},
            "6": {"name": "Aqua", "col": [0, 255, 255]},
            "7": {"name": "Blue", "col": [0, 0, 255]},
            "8": {"name": "Purple", "col": [128, 0, 128]},
            "9": {"name": "Pink", "col": [255, 0, 255]}
        }

        if old_colours == old_colour_json:
            os.remove(path.colours)
        else:
            # Migrate colours from RGB lists to hex strings.
            new_colours = []
            if type(old_colours) != list:
                old_ids = list(old_colours.keys())
                old_ids.sort()
                for uuid in old_ids:
                    new_name = old_colours[uuid]["name"]
                    new_hex = common.rgb_to_hex(old_colours[uuid]["col"])
                    new_colours.append({"name": new_name, "hex": new_hex})

                save_file(path.colours, new_colours)

    # v0.6.0 (dev)
    if config_version < 8:
        from . import effects

        # Convert old 'application profiles' to new 'sequence' effect
        dbg.stdout("Upgrading v0.3.12 application profiles to new format...", dbg.action)
        if os.path.exists(common.paths.old_profile_folder):
            files = glob.glob(common.paths.old_profile_folder + "/*.json")
            for profile_path in files:
                try:
                    with open(profile_path, "r") as f:
                        olddata = json.load(f)
                except Exception:
                    dbg.stdout("Skipping: {0} (Failed to read JSON)".format(profile_path), dbg.error)
                    continue

                # Validate all required fields are present
                required = ["name", "icon", "rows"]
                if not all(key in olddata.keys() for key in required):
                    dbg.stdout("Skipping: {0} (Invalid data)".format(profile_path), dbg.error)
                    continue

                # Prepare new format
                _fx_i18n = locales.Locales("polychromatic")
                _fx__ = _fx_i18n.init()
                fileman = effects.EffectFileManagement()
                newdata = fileman.init_data(olddata["name"], effects.TYPE_SEQUENCE)

                # Migrate data from old format
                # v0.3.12 was hardcoded to BlackWidow Chroma
                newdata["icon"] = olddata["icon"]
                newdata["map_device"] = "Razer BlackWidow Chroma"
                newdata["map_device_icon"] = "keyboard"
                newdata["map_graphic"] = "blackwidow_m_keys_en_GB.svg" if _fx_i18n.get_current_locale() == "en_GB" else "blackwidow_m_keys_en_US.svg"
                newdata["map_cols"] = 22
                newdata["map_rows"] = 6
                newdata["loop"] = False

                # "rows": {y: [[r,g,b], []} => "frames": [{x: {y: "#RRGGBB"}}]
                newdata["frames"].append({})
                for x in range(0, 22):
                    newdata["frames"][0][str(x)] = {}

                for row in olddata["rows"]:
                    for col, rgb in enumerate(olddata["rows"][row]):
                        if not type(rgb) == list or rgb == [0, 0, 0]:
                            continue
                        newdata["frames"][0][str(col)][str(row)] = common.rgb_to_hex(rgb)

                fileman.save_item(newdata)
                dbg.stdout("OK: " + newdata["name"], dbg.success)

        # Rename folders
        if os.path.exists(common.paths.old_profile_folder):
            os.rename(common.paths.old_profile_folder, os.path.join(common.paths.config, "profiles-v0.3.12.old"))

        if os.path.exists(common.paths.old_profile_backups):
            os.rename(common.paths.old_profile_backups, os.path.join(common.paths.config, "backups-v0.3.12.old"))

    # Write new version number.
    data = load_file(path.preferences)
    data["config_version"] = VERSION
    save_file(path.preferences, data)

    dbg.stdout("Configuration successfully upgraded.", dbg.success)


def get_colour_list(_):
    """
    Returns the user's saved colour list, validated for consistency.
    Format: [{"name": <str>, "hex": <str>}, {...}]
    """
    colours = []
    colour_list = load_file(path.colours)

    if not type(colour_list) == list:
        dbg.stdout("colours.json is malformed! Will backup and reset.", dbg.error)
        os.renames(path.colours, path.colours + ".bak")
        init(_)
        colour_list = load_file(path.colours)

    for colour in colour_list:
        try:
            name = colour["name"]
            value = colour["hex"]
            if not common.validate_hex(value):
                raise KeyError
            colours.append(colour)
        except (KeyError, TypeError):
            dbg.stdout("colours.json: Invalid item: {0}".format(str(colour)), dbg.error)
            continue

    return colours


def get_custom_icons():
    """
    Returns a list of all the icons currently stored in the user's "custom icons"
    folder. This is used by the icon picker. Save data will reference images
    by a relative file name.
    """
    return os.listdir(path.custom_icons)


def init(_):
    """
    Prepares the preferences module.
    """
    # Create preferences if non-existent.
    for json_path in [path.preferences]:
        if not os.path.exists(json_path):
            init_config(json_path)

    # Check the configuration and software version matches.
    upgrade_old_pref()

    # Validate preferences with defaults if non-existant.
    data = load_file(path.preferences)

    # Generate colours with defaults if non-existant.
    data = load_file(path.colours)
    if len(data) <= 2:
        default_data = [
            {"name": _("White"), "hex": "#FFFFFF"},
            {"name": _("Red"), "hex": "#FF0000"},
            {"name": _("Green"), "hex": "#00FF00"},
            {"name": _("Blue"), "hex": "#0000FF"},
            {"name": _("Aqua"), "hex": "#00FFFF"},
            {"name": _("Orange"), "hex": "#FFA500"},
            {"name": _("Pink"), "hex": "#FF00FF"},
            {"name": _("Purple"), "hex": "#8000FF"},
            {"name": _("Yellow"), "hex": "#FFFF00"},
            {"name": _("Light Grey"), "hex": "#C0C0C0"},
            {"name": _("Dark Grey"), "hex": "#7F7F7F"},
            {"name": _("Black"), "hex": "#000000"}
        ]
        save_file(path.colours, default_data)



#!/usr/bin/python3
##
## Copyright (C) 2013 Riverbank Computing Limited.
## Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
## All rights reserved.
##
## This file is part of the examples of PyQt.
##
## $QT_BEGIN_LICENSE:BSD$
## You may use this file under the terms of the BSD license as follows:
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are
## met:
##   * Redistributions of source code must retain the above copyright
##     notice, this list of conditions and the following disclaimer.
##   * Redistributions in binary form must reproduce the above copyright
##     notice, this list of conditions and the following disclaimer in
##     the documentation and/or other materials provided with the
##     distribution.
##   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
##     the names of its contributors may be used to endorse or promote
##     products derived from this software without specific prior written
##     permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
## $QT_END_LICENSE$
##
#############################################################################

from PyQt5.QtCore import QPoint, QRect, QSize, Qt
from PyQt5.QtWidgets import QLayout, QSizePolicy, QWidget


class FlowLayout(QLayout):
    def __init__(self, parent=None, margin=0, spacing=-1):
        super(FlowLayout, self).__init__(parent)

        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)

        self.setSpacing(spacing)

        self.itemList = []

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item):
        self.itemList.append(item)

    def count(self):
        return len(self.itemList)

    def itemAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList[index]

        return None

    def takeAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList.pop(index)

        return None

    def expandingDirections(self):
        return Qt.Orientations(Qt.Orientation(0))

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self.doLayout(QRect(0, 0, width, 0), True)
        return height

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()

        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())

        margin, _, _, _ = self.getContentsMargins()

        size += QSize(2 * margin, 2 * margin)
        return size

    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0

        for item in self.itemList:
            wid = item.widget()
            spaceX = self.spacing() + wid.style().layoutSpacing(QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Horizontal)
            spaceY = self.spacing() + wid.style().layoutSpacing(QSizePolicy.PushButton, QSizePolicy.PushButton, Qt.Vertical)
            nextX = x + item.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0

            if not testOnly:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))

            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())

        return y + lineHeight - rect.y()



#!/usr/bin/python3
#
# Unit test runner logic. Consider using ./run.sh instead.
#

import sys
import unittest

import internals
import effects
import fx
import middleman

loader = unittest.TestLoader()
suite  = unittest.TestSuite()

suite.addTests(loader.loadTestsFromModule(internals))
suite.addTests(loader.loadTestsFromModule(effects))
suite.addTests(loader.loadTestsFromModule(fx))
suite.addTests(loader.loadTestsFromModule(middleman))

# Initialize runner
runner = unittest.TextTestRunner()

if len(sys.argv) > 1:
    if sys.argv[1] == "--verbose":
        runner = unittest.TextTestRunner(verbosity=3)

result = runner.run(suite)

if result.wasSuccessful():
    sys.exit(0)

sys.exit(1)



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
Handles the processing of custom software effects and device mapping.
"""

import json
import os
import shutil
import platform

from . import common
from . import fileman

# Effect Types
TYPE_LAYERED = 1
TYPE_SCRIPTED = 2
TYPE_SEQUENCE = 3

# Layer Types
LAYER_STATIC = 10
LAYER_GRADIENT = 11
LAYER_PULSING = 12
LAYER_WAVE = 13
LAYER_SPECTRUM = 14
LAYER_CYCLE = 15
LAYER_SCRIPT = 16


class EffectFileManagement(fileman.FlatFileManagement):
    """
    Provides common functions for parsing the custom effects.
    """
    def __init__(self):
        self.feature = "effects"
        self.factory_path = os.path.join(self.paths.data_dir, "effects")
        self.local_path = self.paths.effects

    def get_item(self, path):
        """
        Load the effect into memory and validate the data is consistent and in
        accordance to the type of effect it is.

        Returns:
            {}          Data (and its effect type specific data) as defined by the documentation.
            ERROR_*     One of ERROR_* variables from the fileman module.
        """
        data = self._load_file(path)
        if not data:
            return fileman.ERROR_MISSING_FILE

        # Check the format and upgrade if necessary.
        try:
            save_format = data["save_format"]
            if save_format > fileman.VERSION:
                self.dbg.stdout("Refusing to load the effect '{0}' as it was created in a newer version of the application.".format(data["name"]), self.dbg.error)
                return fileman.ERROR_NEWER_FORMAT
            elif fileman.VERSION > save_format:
                data = self.upgrade_item(data)
                self.dbg.stdout("Effect upgraded (in memory) from v{0} to v{1}: {2}".format(save_format, fileman.VERSION, path), self.dbg.success, 1)
        except KeyError:
            self.dbg.stdout("Invalid Effect: Unspecified save format!", self.dbg.error)
            return fileman.ERROR_BAD_DATA

        # Validate common metadata
        if not self._validate_key(data, "type", int):
            self.dbg.stdout("Invalid Effect: Unspecified type!", self.dbg.error)
            return fileman.ERROR_BAD_DATA

        results = [
            self._validate_key(data, "name", str),
            self._validate_key(data, "type", int),
            self._validate_key(data, "author", str),
            self._validate_key(data, "icon", str),
            self._validate_key(data, "summary", str),
            self._validate_key(data, "map_device", str),
            self._validate_key(data, "map_device_icon", str),
            self._validate_key(data, "map_graphic", str),
            self._validate_key(data, "map_cols", int),
            self._validate_key(data, "map_rows", int),
            self._validate_key(data, "save_format", int),
            self._validate_key(data, "revision", int)
        ]

        # Validate specific data
        effect_type = data["type"]

        if effect_type == TYPE_LAYERED:
            results.append(self._validate_key(data, "layers", list))
            try:
                for layer in data["layers"]:
                    results.append(self._validate_key(layer, "name", str))
                    results.append(self._validate_key(layer, "type", int))
                    results.append(self._validate_key(layer, "positions", list))
                    results.append(self._validate_key(layer, "properties", dict))
            except KeyError:
                results.append(False)

        elif effect_type == TYPE_SCRIPTED:
            script_path = os.path.join(os.path.dirname(path), os.path.basename(path).replace(".json", ".py"))
            if not os.path.exists(script_path):
                self.dbg.stdout("Effect script does not exist: " + script_path, self.dbg.warning, 1)

            results.append(self._validate_key(data, "required_os", list))
            results.append(self._validate_key(data, "parameters", list))
            results.append(self._validate_key(data, "designed_for", list))
            results.append(self._validate_key(data, "optimised_for", list))
            try:
                for param in data["parameters"]:
                    results.append(self._validate_key(param, "var", str))
                    results.append(self._validate_key(param, "label", str))
                    results.append(self._validate_key(param, "type", str))
                    results.append(self._validate_key(param, "value"))
                    results.append(self._validate_key(param, "default"))
            except KeyError:
                results.append(False)

        elif effect_type == TYPE_SEQUENCE:
            results.append(self._validate_key(data, "fps", int))
            results.append(self._validate_key(data, "loop", bool))
            results.append(self._validate_key(data, "frames", list))

        # Was validation successful?
        if False in results:
            self.dbg.stdout("The effect '{0}' contains invalid data.".format(path), self.dbg.error)
            if not self.dbg.verbose_level >= 1:
                self.dbg.stdout("Run this application in the Terminal with the -v parameter for details.", self.dbg.warning)
            return fileman.ERROR_BAD_DATA

        # Append "parsed" key (used by the UI, but not saved)
        data["parsed"] = self._get_parsed_keys(data, path)

        return data

    def init_data(self, effect_name, effect_type):
        """
        Creates new effect data, ready for editing by the editor.

        Returns (dict) containing new data.
        """
        data = {}

        # Common for all effects
        data["name"] = effect_name
        data["type"] = effect_type
        data["author"] = ""
        data["icon"] = "img/general/effects.svg"
        data["summary"] = ""
        data["map_device"] = ""
        data["map_device_icon"] = ""
        data["map_graphic"] = ""
        data["map_cols"] = 0
        data["map_rows"] = 0
        data["save_format"] = fileman.VERSION
        data["revision"] = 1

        if effect_type == TYPE_LAYERED:
            data["layers"] = [
                {
                    "name": self._("Layer 1"),
                    "type": LAYER_STATIC,
                    "positions": [],
                    "properties": {}
                }
            ]

        elif effect_type == TYPE_SCRIPTED:
            data["required_os"] = []
            data["parameters"] = []
            data["designed_for"] = []
            data["optimised_for"] = []

        elif effect_type == TYPE_SEQUENCE:
            data["fps"] = 10
            data["loop"] = True
            data["frames"] = []

        return data

    def upgrade_item(self, data):
        """
        Upgrades the data for an effect if it was saved in a older version
        of the application.

        Returns the new data.
        """
        old_ver = data["save_format"]

        # Version 8: First! Nothing new yet.

        data["save_format"] = fileman.VERSION
        return data

    def delete_item(self, path):
        """
        In addition to the usual deletion of an item, also delete the
        effect's accompanying script (if a scripted effect)
        """
        data = self._load_file(path)

        if data["type"] == TYPE_SCRIPTED:
            py_path = path.replace(".json", ".py")
            if os.path.exists(py_path):
                os.remove(py_path)
                self.dbg.stdout("Deleted: " + path, self.dbg.success, 1)
            else:
                self.dbg.stdout("Accompanying script file no longer exists: " + path, self.dbg.warning, 1)

        return super().delete_item(path)

    def clone_item(self, path):
        """
        In addition to the usual duplication of an item, also copy
        the effect's accompanying script (if a scripted effect)

        Returns:
            (str)           Success: Path to the new effect
            None            Failed to clone effect
        """
        new_path = super().clone_item(path)

        if new_path:
            data = self._load_file(new_path)

            if data["type"] == TYPE_SCRIPTED:
                src_script = path.replace(".json", ".py")
                dest_script = new_path.replace(".json", ".py")

                if not os.path.exists(src_script):
                    return False

                shutil.copy(src_script, dest_script)
                self.dbg.stdout("Clone OK: " + dest_script, self.dbg.success)

            return new_path

        return None


class DeviceMapGraphics(object):
    """
    Responsible for populating the list of device graphics and generating SVG
    graphics of a grid if a specific device is unavailable (or by request).
    """
    def __init__(self, appdata):
        self.appdata = appdata
        self.map_index = os.path.join(common.paths.data_dir, "devicemaps", "maps.json")
        self.map_dir = os.path.join(common.paths.data_dir, "devicemaps")
        self.cache_dir = common.paths.assets_cache

    def get_graphic_list(self):
        """
        Returns a list of dictionaries referencing graphics:
        {
            "Human readable name": {
                "filename": "some_device_graphic_en_GB.svg",
                "rows": 1,
                "cols": 2,
                "locale": "en_GB"
            }, { ... }
        }
        """
        with open(self.map_index) as f:
            original_svg =  json.load(f)

        parsed_svg = {}

        for name in original_svg:
            graphic_path = self.get_graphic_path(original_svg[name]["filename"])
            if not os.path.exists(graphic_path):
                self.appdata.dbg.stdout("Graphic missing: " + graphic_path, self.appdata.dbg.warning)
                continue
            parsed_svg[name] = original_svg[name]

        return parsed_svg

    def get_graphic_path(self, filename):
        """
        Returns an absolute path to the graphic.
        For use with metadata editor UI which loads via file.
        """
        return os.path.join(self.map_dir, filename)

    def get_grid_path(self, cols, rows):
        """
        Returns an absolute path to the grid SVG.
        For use with metadata editor UI which loads via file.
        """
        svg = self.get_svg_grid(cols, rows)
        svg_path = os.path.join(self.cache_dir, "grid-{0}-{1}.svg".format(cols, rows))
        with open(svg_path, "w") as f:
            f.writelines(svg)
        return svg_path

    def get_graphic_name_from_filename(self, filename):
        """
        Return the human friendly name from the specified filename. If there
        isn't an entry, the filename will be returned.
        """
        maps = self.get_graphic_list()
        for name in maps.keys():
            if maps[name]["filename"] == filename:
                return name
        return filename

    def get_svg_graphic(self, filename):
        """
        Loads the SVG of a device to visually map.

        Returns:
            (str)       SVG Data
            None        File Missing
        """
        # Verify the device map exists, then load it
        graphic_path = os.path.join(self.map_dir, filename)

        if not os.path.exists(graphic_path):
            return None

        with open(os.path.join(self.map_dir, filename)) as f:
            return " ".join(f.readlines()).replace("\\", "\\\\").replace("\n", "")

    def get_svg_grid(self, cols, rows):
        """
        Returns the SVG of the device's matrix represented as a 'pretty' graphic.
        """
        svg = []

        # How large is each grid?
        square_px = 50
        margin_px = 1
        fill_colour = "#00ff00"
        stroke_colour = "#008000"
        stroke_width = 1
        total_X_blocks = cols
        total_Y_blocks = rows

        svg.append('<svg width="{width}px" height="{height}px"> version="1.1" viewBox="0px 0px {width}px {height}px" xmlns="http://www.w3.org/2000/svg">'.format(
            width = total_X_blocks * square_px + (margin_px * total_X_blocks),
            height = total_Y_blocks * square_px + (margin_px * total_X_blocks)
        ))

        for x in range(0, total_X_blocks):
            for y in range(0, total_Y_blocks):
                x_pos = x * square_px + (x * margin_px + 1)
                y_pos = y * square_px + (y * margin_px + 1)
                svg.append('<g id="x{x}-y{y}" class="LED"><rect x="{x_pos}px" y="{y_pos}px" width="{square_px}px" height="{square_px}px" style="fill:{fill_colour};paint-order:markers fill stroke;stroke-linecap:round;stroke-width:{stroke_width};stroke:{stroke_colour}"/></g>'.format(
                    x = x,
                    y = y,
                    x_pos = x_pos,
                    y_pos = y_pos,
                    square_px = square_px,
                    fill_colour = fill_colour,
                    stroke_colour = stroke_colour,
                    stroke_width = stroke_width
                ))

        svg.append("</svg>")
        return "".join(svg)
